import abc
import logging
import math
from os import path, listdir
from pathlib import Path
from typing import List, Dict, Tuple, Generator

import pandas as pd

from genomics_data_index.storage.io.SampleDataPackage import SampleDataPackage
from genomics_data_index.storage.io.SampleDataPackageFactory import SampleDataPackageFactory
from genomics_data_index.storage.io.SampleFilesProcessor import SampleFilesProcessor
from genomics_data_index.storage.io.mutation.NucleotideSampleDataPackage import NucleotideSampleDataPackage
from genomics_data_index.storage.io.mutation.variants_processor.MultipleProcessVcfVariantsTableProcessor import \
    MultipleProcessVcfVariantsTableProcessorFactory
from genomics_data_index.storage.io.mutation.variants_processor.SerialVcfVariantsTableProcessor import \
    SerialVcfVariantsTableProcessorFactory
from genomics_data_index.storage.io.mutation.variants_processor.VcfVariantsTableProcessor import \
    VcfVariantsTableProcessorFactory
from genomics_data_index.storage.io.processor.MultipleProcessSampleFilesProcessor import \
    MultipleProcessSampleFilesProcessor
from genomics_data_index.storage.io.processor.SerialSampleFilesProcessor import SerialSampleFilesProcessor
from genomics_data_index.storage.util.ListSliceIter import ListSliceIter
from genomics_data_index.storage.util.SamplesProgressLogger import SamplesProgressLogger

logger = logging.getLogger(__file__)


class NucleotideSampleDataPackageFactory(SampleDataPackageFactory, abc.ABC):

    def __init__(self, ncores: int, index_unknown: bool, preprocess_dir: Path):
        super().__init__()
        self._ncores = ncores
        self._preprocess_dir = preprocess_dir
        self._index_unknown = index_unknown
        self._sample_files_processor, self._variants_processor_factory = self._create_file_processors()

    @abc.abstractmethod
    def number_samples(self) -> int:
        pass

    def number_batches(self, batch_size: int) -> int:
        return (int)(math.ceil(self.number_samples() / batch_size))

    def _create_file_processors(self) -> Tuple[SampleFilesProcessor, VcfVariantsTableProcessorFactory]:
        if self._ncores > 1:
            file_processor = MultipleProcessSampleFilesProcessor(preprocess_dir=Path(self._preprocess_dir),
                                                                 processing_cores=self._ncores)
            variants_processor_factory = MultipleProcessVcfVariantsTableProcessorFactory(ncores=self._ncores)
        else:
            file_processor = SerialSampleFilesProcessor(preprocess_dir=self._preprocess_dir)
            variants_processor_factory = SerialVcfVariantsTableProcessorFactory.instance()

        return file_processor, variants_processor_factory


class NucleotideInputFilesSampleDataPackageFactory(NucleotideSampleDataPackageFactory):

    def __init__(self, ncores: int, index_unknown: bool, preprocess_dir: Path,
                 input_files_file: Path):
        super().__init__(ncores=ncores, index_unknown=index_unknown, preprocess_dir=preprocess_dir)
        self._input_files_file = input_files_file
        self._sample_vcf, self._mask_files = self.create_sample_vcf_mask(input_files_file)

    def number_samples(self) -> int:
        return len(self._sample_vcf)

    def expected_input_columns(self) -> List[str]:
        return ['Sample', 'VCF', 'Mask File']

    def create_sample_vcf_mask(self, input_files_file) -> Tuple[Dict[str, Path], Dict[str, Path]]:
        logger.debug(f'Examining files listed in {input_files_file}')
        sample_vcf_map = {}
        mask_files_map = {}
        files_df = pd.read_csv(input_files_file, sep='\t')
        for index, row in files_df.iterrows():
            if row['Sample'] in sample_vcf_map:
                raise Exception(f'Error, duplicate samples {row["Sample"]} in file {input_files_file}')

            sample_vcf_map[row['Sample']] = Path(row['VCF'])
            if not pd.isna(row['Mask File']):
                mask_files_map[row['Sample']] = Path(row['Mask File'])

        logger.info(f'Found {len(sample_vcf_map)} VCFs and {len(mask_files_map)} mask files in [{input_files_file}]')

        return sample_vcf_map, mask_files_map

    def create_data_package_iter(self, batch_size: int = 100) -> Generator[SampleDataPackage, None, None]:
        sample_names = list(self._sample_vcf.keys())
        number_batches = self.number_batches(batch_size)
        SamplesProgressLogger.set_total_batches(number_batches)
        sample_names_iter = ListSliceIter(sample_names, batch_size)
        current_batch = 1
        for sample_names_slice in sample_names_iter.islice():
            sample_vcfs_slice = {s: self._sample_vcf[s] for s in sample_names_slice}
            mask_files_slice = {s: self._mask_files[s] for s in sample_names_slice if s in self._mask_files}
            SamplesProgressLogger.set_current_batch(current_batch)
            logger.info(f'Starting batch {current_batch}/{number_batches}')
            yield self._create_data_package(sample_vcfs=sample_vcfs_slice, sample_mask_files=mask_files_slice)
            logger.info(f'Finished batch {current_batch}/{number_batches}')
            current_batch = current_batch + 1
        SamplesProgressLogger.unset_batch_mode()
        logger.debug(f'Finished creating data packages for all {self.number_samples()} samples')

    def _create_data_package(self, sample_vcfs: Dict[str, Path],
                             sample_mask_files: Dict[str, Path]) -> SampleDataPackage:
        return NucleotideSampleDataPackage.create_from_vcf_masks(sample_vcf_map=sample_vcfs,
                                                                 masked_genomic_files_map=sample_mask_files,
                                                                 sample_files_processor=self._sample_files_processor,
                                                                 variants_processor_factory=self._variants_processor_factory,
                                                                 index_unknown_missing=self._index_unknown)

    def create_data_package(self) -> SampleDataPackage:
        sample_vcf, mask_files = self.create_sample_vcf_mask(self._input_files_file)
        return self._create_data_package(sample_vcfs=sample_vcf, sample_mask_files=mask_files)


class NucleotideSnippySampleDataPackageFactory(NucleotideSampleDataPackageFactory):

    def __init__(self, ncores: int, index_unknown: bool, preprocess_dir: Path,
                 snippy_dir: Path):
        super().__init__(ncores=ncores, index_unknown=index_unknown, preprocess_dir=preprocess_dir)
        self._snippy_dir = snippy_dir
        self._sample_dirs = [snippy_dir / d for d in listdir(snippy_dir) if path.isdir(snippy_dir / d)]

    def number_samples(self) -> int:
        return len(self._sample_dirs)

    def create_data_package_iter(self, batch_size: int = 100) -> Generator[SampleDataPackage, None, None]:
        number_batches = self.number_batches(batch_size)
        SamplesProgressLogger.set_total_batches(number_batches)
        sample_dirs_iter = ListSliceIter(self._sample_dirs, batch_size)
        current_batch = 1
        for sample_dirs in sample_dirs_iter.islice():
            SamplesProgressLogger.set_current_batch(current_batch)
            logger.info(f'Starting batch {current_batch}/{number_batches}')
            yield self._create_data_package(sample_dirs=sample_dirs)
            logger.info(f'Finished batch {current_batch}/{number_batches}')
            current_batch = current_batch + 1
        SamplesProgressLogger.unset_batch_mode()
        logger.debug(f'Finished creating data packages for all {self.number_samples()} samples')

    def _create_data_package(self, sample_dirs: List[Path]) -> SampleDataPackage:
        data_package = NucleotideSampleDataPackage.create_from_snippy(sample_dirs=sample_dirs,
                                                                      sample_files_processor=self._sample_files_processor,
                                                                      variants_processor_factory=self._variants_processor_factory,
                                                                      index_unknown_missing=self._index_unknown)
        return data_package

    def create_data_package(self) -> SampleDataPackage:
        logger.debug(
            f'Found {len(self._sample_dirs)} directories in snippy_dir=[{self._snippy_dir}], loading files from '
            f'these directories.')
        return self._create_data_package(sample_dirs=self._sample_dirs)



import re
from pathlib import Path

import pandas as pd

from genomics_data_index.storage.io.mlst.MLSTFeaturesReader import MLSTFeaturesReader
from genomics_data_index.storage.model import MLST_UNKNOWN_ALLELE


class MLSTChewbbacaReader(MLSTFeaturesReader):

    def __init__(self, mlst_file: Path, scheme: str):
        super().__init__()

        self._mlst_file = mlst_file

        if scheme is None:
            raise Exception('scheme cannot be None')

        self._scheme = scheme

    def _read_features_table(self) -> pd.DataFrame:
        df = pd.read_csv(self._mlst_file, sep='\t', dtype=str)
        df = df.rename(columns={
            'FILE': 'File',
        })

        df['Sample'] = self._get_sample_from_filename(df['File'])
        df = self._extract_locus_alleles(df)
        df['Scheme'] = self._scheme

        df = df[['File', 'Sample', 'Scheme', 'Locus', 'Allele']].sort_values(
            by=['Sample', 'Scheme', 'Locus']).reset_index().drop(columns='index')

        return df

    def _is_valid_allele(self, allele: str) -> bool:
        return allele != MLST_UNKNOWN_ALLELE and bool(re.match(r'^\d+$', allele))

    def _get_sample_from_filename(self, filename_series: pd.Series) -> pd.Series:
        file_sample_name_regex = r'^([^.]*)'
        return filename_series.str.extract(file_sample_name_regex, expand=True)

    def _extract_locus_alleles(self, df: pd.DataFrame) -> pd.DataFrame:
        locus_allele_list = list(set(df.columns) - {'File', 'Sample'})
        df = df.melt(id_vars=['File', 'Sample'], value_vars=locus_allele_list,
                     var_name='Locus', value_name='Allele')
        df['Allele'] = df['Allele'].astype(str)
        return df



import logging
import subprocess
from typing import List

logger = logging.getLogger(__name__)

TRACE_LEVEL = 5


def execute_commands(commands: List[List[str]]):
    try:
        for command in commands:
            logger.log(TRACE_LEVEL, f'Running [{" ".join(command)}]')
            subprocess.run(command, stdout=subprocess.PIPE, stderr=subprocess.PIPE, check=True, text=True)
    except subprocess.CalledProcessError as e:
        err_msg = str(e.stderr.strip())
        raise Exception(f'Could not run [{" ".join(e.cmd)}]: error {err_msg}')


def execute_command(command: List[str], stdout=subprocess.PIPE):
    try:
        logger.log(TRACE_LEVEL, f'Running [{" ".join(command)}]')
        subprocess.run(command, stdout=stdout, stderr=subprocess.PIPE, check=True, text=True)
    except subprocess.CalledProcessError as e:
        err_msg = str(e.stderr.strip())
        raise Exception(f'Could not run [{" ".join(e.cmd)}]: error {err_msg}')



from __future__ import annotations

import tempfile
from pathlib import Path
from typing import List, Set, Dict, Generator, Tuple

import pandas as pd
from Bio import SeqIO
from Bio.SeqRecord import SeqRecord
from pybedtools import BedTool

from genomics_data_index.storage.model import NUCLEOTIDE_UNKNOWN, NUCLEOTIDE_UNKNOWN_TYPE


class MaskedGenomicRegions:

    def __init__(self, mask: BedTool):
        self._mask = mask.sort().merge()

    @property
    def mask(self):
        return self._mask

    def intersect(self, other: MaskedGenomicRegions) -> MaskedGenomicRegions:
        return MaskedGenomicRegions(self._mask.intersect(other._mask))

    def subtract(self, other: MaskedGenomicRegions) -> MaskedGenomicRegions:
        subtraction = self._mask.subtract(other._mask)
        return MaskedGenomicRegions(subtraction)

    def union(self, other: MaskedGenomicRegions) -> MaskedGenomicRegions:
        union = self._mask.cat(other._mask, postmerge=True, force_truncate=True)
        return MaskedGenomicRegions(union)

    def mask_to_features(self) -> pd.DataFrame:
        mask_features = []
        ref = 1
        alt = NUCLEOTIDE_UNKNOWN
        nuc_type = NUCLEOTIDE_UNKNOWN_TYPE
        for sequence_name, position in self.positions_iter(start_position_index='1'):
            variant_id = f'{sequence_name}:{position}:{ref}:{alt}'
            mask_features.append([sequence_name, position, ref, alt, nuc_type, variant_id])

        return pd.DataFrame(mask_features, columns=['CHROM', 'POS', 'REF', 'ALT', 'TYPE', 'VARIANT_ID'])

    def mask_genome(self, genome_file: Path, mask_char: str = '?', remove: bool = True) -> Dict[str, SeqRecord]:
        """
        Gets a SeqRecord with all those regions on the passed genome that are in the masked regions removed
        (or masked with mask_char).
        :param genome_file: The genome file to mask.
        :param mask_char: The character to mask with.
        :param remove: Whether or not to remove masked sequence data.
        :return: A Dictionary mapping a sequence name to a SeqRecord containing all those regions on the sequence
                 within the masked regions removed (or masked with mask_char)
        """
        with tempfile.TemporaryDirectory() as out_f:
            seq_records = {}
            output_fasta = Path(out_f) / 'masked.fasta'
            self._mask.mask_fasta(fi=str(genome_file), fo=str(output_fasta), mc=mask_char)
            for record in SeqIO.parse(output_fasta, 'fasta'):
                if remove:
                    record.seq = record.seq.ungap(mask_char)
                seq_records[record.id] = record
            return seq_records

    def write(self, file: Path):
        self._mask.saveas(str(file), compressed=True)

    @classmethod
    def union_all(cls, masked_regions: List[MaskedGenomicRegions]):
        if len(masked_regions) == 0:
            raise Exception('Cannot merge empty list')
        elif len(masked_regions) == 1:
            return masked_regions[0]
        else:
            start_mask = masked_regions.pop()
            union = start_mask._mask.cat(*[o._mask for o in masked_regions], postmerge=True, force_truncate=True)
            return MaskedGenomicRegions(union)

    @classmethod
    def from_sequences(cls, sequences: List[SeqRecord]) -> MaskedGenomicRegions:
        def is_missing(char):
            return char.upper() == 'N' or char == '-'

        # pybedtools internally stores as 0-based BED file intervals
        # https://daler.github.io/pybedtools/intervals.html#bed-is-0-based-others-are-1-based
        mask_intervals = []

        for record in sequences:
            start = 0
            in_mask = False
            for idx, char in enumerate(record.seq):
                if in_mask:
                    if not is_missing(char):
                        in_mask = False
                        # pybedtools stop position is not included in interval
                        stop = idx
                        mask_intervals.append((record.id, start, stop))
                else:
                    if is_missing(char):
                        in_mask = True
                        start = idx

            # Finish recording last interval if it exists (e.g., if last bit of sequence was like 'NNNN')
            if in_mask:
                stop = len(record)
                mask_intervals.append((record.id, start, stop))

        bedtool_intervals = BedTool(mask_intervals)
        return MaskedGenomicRegions(bedtool_intervals)

    @classmethod
    def from_file(cls, file: Path) -> MaskedGenomicRegions:
        bed_file_data = BedTool(str(file))
        return MaskedGenomicRegions(bed_file_data)

    @classmethod
    def from_vcf_file(cls, file: Path) -> MaskedGenomicRegions:
        bed_file_data = BedTool(str(file)).merge()
        return MaskedGenomicRegions(bed_file_data)

    @classmethod
    def empty_mask(cls):
        return MaskedGenomicRegions(BedTool('', from_string=True))

    def is_empty(self):
        return len(self) == 0

    def sequence_names(self) -> Set[str]:
        """
        Gets a set of sequence names from this genomic regions mask.
        :return: A set of sequence names.
        """
        return {x.chrom for x in self._mask}

    def contains(self, sequence: str, position: int, start_position_index: str = '0') -> bool:
        if start_position_index != '0' and start_position_index != '1':
            raise Exception((f'Unknown value start_position_index=[{start_position_index}].'
                             'Should be "0" or "1" to indicate which is the starting base position'))
        elif start_position_index == '1':
            position = position - 1

        for i in self._mask:
            if i.chrom == sequence and i.start <= position < i.end:
                return True
        return False

    def _validate_start_position_index(self, start_position_index: str) -> None:
        if start_position_index not in ['0', '1']:
            raise Exception((f'Unknown value start_position_index=[{start_position_index}].'
                             'Should be "0" or "1" to indicate which is the starting base position'))

    def overlaps_range(self, sequence: str, start: int, stop: int, start_position_index: str = '0') -> bool:
        self._validate_start_position_index(start_position_index)

        if start_position_index == '1':
            start = start - 1
            stop = stop - 1

        if stop <= start:
            raise Exception(f'start=[{start}] is less than stop=[{stop}]')

        for i in self._mask:
            if i.chrom == sequence:
                if i.start <= start and i.end > start:
                    return True
                elif start < i.end and stop > i.end:
                    return True
        return False

    def positions_iter(self, start_position_index: str = '0') -> Generator[Tuple[str, int], None, None]:
        """
        Creates an iterator to iterate over all ('sequence', 'position') in this mask.
        :param start_position_index: Whether positions should be in 0-base coordinates or 1-base coordinates.
                                     See https://bedtools.readthedocs.io/en/latest/content/general-usage.html#bed-format
                                     for a description of the differences in coordinates.
        :return: An iterator which will return tuples like ('sequence', 'position') for every
                 position in this mask.
        """
        self._validate_start_position_index(start_position_index)

        for sequence in self._mask:
            sequence_name = sequence.chrom
            start = sequence.start
            end = sequence.end

            if start_position_index == '1':
                start = start + 1
                end = end + 1

            for pos in range(start, end):
                yield sequence_name, pos

    def __len__(self) -> int:
        """
        Calculates length of underlying masked intervals. Assumes the intervals have been merged beforehand.
        :return: The length of the masked intervals.
        """
        total = 0
        for i in self._mask:
            total += len(i)
        return total



import abc
import logging
from pathlib import Path
from typing import List, Callable, Any, Union

import pandas as pd
from sqlalchemy import create_engine
from sqlalchemy.orm import sessionmaker

import genomics_data_index.storage.model.db
from genomics_data_index.storage.model.db import Base
from genomics_data_index.storage.model.db.DatabasePathTranslator import DatabasePathTranslator
from genomics_data_index.storage.util import TRACE_LEVEL
from genomics_data_index.storage.util.ListSliceIter import ListSliceIter

logger = logging.getLogger(__name__)


class DatabaseConnection:

    def __init__(self, connection_string: str, database_path_translator: DatabasePathTranslator):
        self._setup_database_type(connection_string)

        engine = create_engine(connection_string, echo=False)

        Session = sessionmaker(bind=engine)
        self._session = Session()
        self._database_path_translator = database_path_translator

        # Sets global variable here for translating between relative/absolute paths in the database
        # I don't like using global variables and have to look into some other method to set this later
        if genomics_data_index.storage.model.db.database_path_translator is not None:
            logger.warning(f'Attempting to set global database_path_translator={database_path_translator}'
                           ' but it is already set')
        genomics_data_index.storage.model.db.database_path_translator = database_path_translator

        Base.metadata.create_all(engine)

    def _setup_database_type(self, connection_string: str) -> None:
        logger.debug(f'Setup database type: {connection_string}')
        self._is_sqlite = connection_string.startswith('sqlite')

        if self._is_sqlite:
            self._sqlite_path = Path(connection_string[len('sqlite:///'):])

    def get_session(self):
        return self._session

    def get_database_size(self) -> pd.DataFrame:
        if self._is_sqlite:
            db_size = self._sqlite_path.stat().st_size
            size_df = pd.DataFrame([['Database', str(self._sqlite_path.name), 'All', db_size, pd.NA, 1]],
                                   columns=['Type', 'Name', 'Division', 'Data Size', 'Index Size', 'Number of Items'])
            return size_df
        else:
            database_name = self._session.bind.url.database

            table_names = self._session.bind.table_names()

            # TODO: This could maybe be improved to use something more appropriate form SQLAlchemy
            statement = f'''
            SELECT table_name AS `Table`,
            data_length as `Data size`,
            index_length as `Index size`
            FROM information_schema.TABLES
            WHERE table_schema = "{database_name}";
            '''
            result = self._session.execute(statement)
            size_of_table = result.fetchall()

            size_df = pd.DataFrame(size_of_table, columns=['Division', 'Data Size', 'Index Size'])
            size_df.insert(0, 'Name', database_name)
            size_df.insert(0, 'Type', 'Database')

            # Get number of rows from tables in database
            size_df['Number of Items'] = pd.NA
            for name in table_names:
                count_statement = f'''
                SELECT count(*) from {name};
                '''
                result = self._session.execute(count_statement)
                table_counts = result.fetchall()
                assert 1 == len(table_counts)

                size_df.loc[size_df['Division'] == name, 'Number of Items'] = table_counts[0][0]

            return size_df


class EntityExistsError(Exception):

    def __init__(self, msg):
        super().__init__(msg)


class SQLQueryInBatcher(abc.ABC):

    def __init__(self, in_data: Union[List[str], List[int]], batch_size: int):
        self._batch_size = batch_size
        self._in_data_size = len(in_data)
        self._in_data_slicer = ListSliceIter(in_data, slice_size=batch_size)

    @abc.abstractmethod
    def _do_update(self, processed_data: Any, slice_processed_data: Any) -> None:
        pass

    @abc.abstractmethod
    def _initialize_processed_data(self) -> Any:
        pass

    def process(self, batch_func: Callable[[Union[List[str], List[int]]], Any]) -> Any:
        processed_data = self._initialize_processed_data()
        slice_number = 0
        logger.debug(f'Dividing up SQL query with IN statement for {self._in_data_size} data elements '
                     f'into a maximum of {self._batch_size} elements per SQL query')
        for in_slice in self._in_data_slicer.islice():
            logger.log(TRACE_LEVEL, f'Processing slice={slice_number}')
            slice_processed_data = batch_func(in_slice)
            self._do_update(processed_data=processed_data,
                            slice_processed_data=slice_processed_data)
            slice_number = slice_number + 1
        logger.debug(f'Finished SQL query with IN statement for {self._in_data_size} data elements.')

        return processed_data


class SQLQueryInBatcherDict(SQLQueryInBatcher):

    def __init__(self, in_data: Union[List[str], List[int]], batch_size: int):
        super().__init__(in_data=in_data, batch_size=batch_size)

    def _initialize_processed_data(self) -> Any:
        return {}

    def _do_update(self, processed_data: Any, slice_processed_data: Any) -> None:
        processed_data.update(slice_processed_data)


class SQLQueryInBatcherList(SQLQueryInBatcher):

    def __init__(self, in_data: Union[List[str], List[int]], batch_size: int):
        super().__init__(in_data=in_data, batch_size=batch_size)

    def _initialize_processed_data(self) -> Any:
        return []

    def _do_update(self, processed_data: Any, slice_processed_data: Any) -> None:
        processed_data.extend(slice_processed_data)



import pytest

from genomics_data_index.storage.model.QueryFeatureHGVS import QueryFeatureHGVS


def test_create_from_id():
    f = QueryFeatureHGVS.create_from_id('hgvs:NC_011083:SEHA_RS01180:c.497C>A')
    assert 'hgvs:NC_011083:SEHA_RS01180:c.497C>A' == f.id
    assert 'NC_011083:SEHA_RS01180:c.497C>A' == f.id_no_prefix
    assert 'NC_011083' == f.sequence
    assert 'NC_011083' == f.scope
    assert 'SEHA_RS01180' == f.gene
    assert 'c.497C>A' == f.variant
    assert f.has_id()
    assert f.has_gene()
    assert f.is_nucleotide()
    assert not f.is_protein()

    f = QueryFeatureHGVS.create_from_id('hgvs:NC_011083:SEHA_RS01180:p.Ala166Glu')
    assert 'hgvs:NC_011083:SEHA_RS01180:p.Ala166Glu' == f.id
    assert 'NC_011083:SEHA_RS01180:p.Ala166Glu' == f.id_no_prefix
    assert 'NC_011083' == f.sequence
    assert 'NC_011083' == f.scope
    assert 'SEHA_RS01180' == f.gene
    assert 'p.Ala166Glu' == f.variant
    assert f.has_id()
    assert f.has_gene()
    assert not f.is_nucleotide()
    assert f.is_protein()

    # Test to make sure error if prefix is incorrect
    with pytest.raises(Exception) as execinfo:
        QueryFeatureHGVS.create_from_id('hgvs_gn:NC_011083:SEHA_RS01180:c.497C>A')

    assert 'hgvs_id=[hgvs_gn:NC_011083:SEHA_RS01180:c.497C>A] must start with [hgvs:]' in str(execinfo)


def test_create():
    f = QueryFeatureHGVS.create('NC_011083', 'SEHA_RS01180', 'c.497C>A')
    assert 'hgvs:NC_011083:SEHA_RS01180:c.497C>A' == f.id
    assert 'NC_011083:SEHA_RS01180:c.497C>A' == f.id_no_prefix
    assert 'NC_011083' == f.sequence
    assert 'NC_011083' == f.scope
    assert 'SEHA_RS01180' == f.gene
    assert 'c.497C>A' == f.variant
    assert f.has_id()
    assert f.has_gene()
    assert f.is_nucleotide()
    assert not f.is_protein()

    f = QueryFeatureHGVS.create('NC_011083', 'SEHA_RS01180', 'p.Ala166Glu')
    assert 'hgvs:NC_011083:SEHA_RS01180:p.Ala166Glu' == f.id
    assert 'NC_011083:SEHA_RS01180:p.Ala166Glu' == f.id_no_prefix
    assert 'NC_011083' == f.sequence
    assert 'NC_011083' == f.scope
    assert 'SEHA_RS01180' == f.gene
    assert 'p.Ala166Glu' == f.variant
    assert f.has_id()
    assert f.has_gene()
    assert not f.is_nucleotide()
    assert f.is_protein()

    f = QueryFeatureHGVS.create('NC_011083', None, 'n.1031571T>C')
    assert 'hgvs:NC_011083:n.1031571T>C' == f.id
    assert 'NC_011083:n.1031571T>C' == f.id_no_prefix
    assert 'NC_011083' == f.sequence
    assert 'NC_011083' == f.scope
    assert f.gene is None
    assert 'n.1031571T>C' == f.variant
    assert f.has_id()
    assert not f.has_gene()
    assert f.is_nucleotide()
    assert not f.is_protein()

    # Don't include gene id if variant starts with 'n.' (coordinates with respect to sequence
    f = QueryFeatureHGVS.create('NC_011083', 'gene', 'n.1031571T>C')
    assert 'hgvs:NC_011083:n.1031571T>C' == f.id
    assert 'NC_011083:n.1031571T>C' == f.id_no_prefix
    assert 'NC_011083' == f.sequence
    assert 'NC_011083' == f.scope
    assert f.gene is None
    assert 'n.1031571T>C' == f.variant
    assert f.has_id()
    assert not f.has_gene()
    assert f.is_nucleotide()
    assert not f.is_protein()

    # None
    f = QueryFeatureHGVS.create('NC_011083', 'gene', None)
    assert f.id is None
    assert f.id_no_prefix is None
    assert 'NC_011083' == f.sequence
    assert 'NC_011083' == f.scope
    assert f.gene == 'gene'
    assert f.variant is None
    assert f.has_gene()
    assert not f.has_id()
    assert not f.is_nucleotide()
    assert not f.is_protein()

    # None2
    f = QueryFeatureHGVS.create('NC_011083', None, None)
    assert f.id is None
    assert f.id_no_prefix is None
    assert 'NC_011083' == f.sequence
    assert 'NC_011083' == f.scope
    assert f.gene is None
    assert f.variant is None
    assert not f.has_gene()
    assert not f.has_id()
    assert not f.is_nucleotide()
    assert not f.is_protein()



import logging
from typing import Union, Iterable, Dict

from ete3 import Tree, TreeStyle, NodeStyle

from genomics_data_index.api.query.SamplesQuery import SamplesQuery
from genomics_data_index.api.viewer.TreeSamplesVisual import TreeSamplesVisual

logger = logging.getLogger(__name__)


class HighlightTreeSamplesVisual(TreeSamplesVisual):
    """
    Class used to highlight the set of samples with a particular color/node style.
    """

    def __init__(self, samples: Union[SamplesQuery, Iterable[str]],
                 present_node_style: NodeStyle,
                 unknown_node_style: NodeStyle,
                 include_unknown: bool,
                 legend_color: str, legend_label: str,
                 legend_nodesize: int,
                 legend_fontsize: int,
                 legend_columns: Dict[str, int]):
        """
        Builds a new HighlightTreeSamplesVisual with the given information.
        :param samples: The set of samples to highlight.
        :param present_node_style: The style of the nodes in the tree for the present set of samples.
        :param unknown_node_style: The style of the ndoes in the tree for the unknown set of samples.
        :param include_unknown: Whether or not unknown samples should be included.
        :param legend_color: The color of the legend item.
        :param legend_label: The legend label.
        :param legend_nodesize: The size of the legend node.
        :param legend_fontsize: The size of the legend font.
        :param legend_columns: A dict mapping legend item types to column numbers.
        :return: A new HighlightTreeSamplesVisual.
        """
        super().__init__(samples=samples, legend_nodesize=legend_nodesize,
                         legend_fontsize=legend_fontsize, legend_columns=legend_columns)
        self._present_node_style = present_node_style
        self._unknown_node_style = unknown_node_style
        self._include_unknown = include_unknown
        self._legend_color = legend_color
        self._legend_label = legend_label
        self._unknown_legend_color = self._unknown_node_style['bgcolor']

    def _get_and_set_style(self, sample_name: str, node_style: NodeStyle, tree: Tree) -> None:
        nodes = tree.get_leaves_by_name(sample_name)
        if len(nodes) == 0:
            logger.warning(f'Could not find sample=[{sample_name}] in tree. Not highlighting.')
        elif len(nodes) > 1:
            raise Exception(f'More than one node in the tree matched sample=[{sample_name}]')
        else:
            node = nodes[0]
            node.set_style(node_style)

    def apply_visual(self, tree: Tree, tree_style: TreeStyle) -> None:
        has_unknowns = self._include_unknown and len(self.unknown_sample_names) > 0
        self._add_legend_entry(self._legend_label, legend_color=self._legend_color,
                               unknown_color=self._unknown_legend_color,
                               include_unknown=self._include_unknown,
                               has_unknowns=has_unknowns,
                               kind='rect', tree_style=tree_style)

        for name in self.present_sample_names:
            self._get_and_set_style(name, node_style=self._present_node_style, tree=tree)

        if self._include_unknown:
            for name in self.unknown_sample_names:
                self._get_and_set_style(name, node_style=self._unknown_node_style, tree=tree)



from __future__ import annotations

import abc
from typing import Union, List, Set, Tuple

import numpy as np
import pandas as pd
from ete3 import Tree

from genomics_data_index.api.query.kind.IsaKind import IsaKind
from genomics_data_index.storage.SampleSet import SampleSet
from genomics_data_index.storage.model.QueryFeature import QueryFeature


class SamplesQuery(abc.ABC):
    """
    The base class for queries related to samples. The different query methods in a SamplesQuery
    operate on a set of samples (specifically a set of Sample identifiers,
    see :py:class:`genomics_data_index.storage.SampleSet`) and returns a new SamplesQuery consisting of the
    subset of samples matching the defined critera.

    That is to say, if A = SamplesQuery(), then B = A.method() will consist of a subset of A matching the criteria
    defined by A.method().

    Multiple subtypes of SamplesQuery exist which may provided access to data joined to a set of Samples for
    implementing different queries (e.g., a :py:class:`genomics_data_index.api.query.impl.TreeSamplesQuery` provides
    a tree joined to a SampleSet and methods to query based on the tree.
    """

    ALL_SAMPLES = SampleSet.create_all()

    def __init__(self):
        pass

    @property
    @abc.abstractmethod
    def universe_set(self) -> SampleSet:
        """
        The set of samples :py:class:`genomics_data_index.storage.SampleSet` defining the universe under consideration
        for this SamplesQuery. Useful for e.g., complement() functionality.

        :returns: A SampleSet.
        """
        pass

    @property
    @abc.abstractmethod
    def sample_set(self) -> SampleSet:
        """
        The set of selected samples :py:class:`genomics_data_index.storage.SampleSet` in this query.

        :returns: A SampleSet defining the selected samples of this query.
        """
        pass

    @property
    @abc.abstractmethod
    def unknown_set(self) -> SampleSet:
        """
        The set of samples :py:class:`genomics_data_index.storage.SampleSet` for which it is unknown
        whether they match this query or not (e.g., due to missing/unknown data on a region of the genome).

        :returns: A SampleSet defining the samples where it is unknown if they match the query or not.
        """
        pass

    @property
    @abc.abstractmethod
    def absent_set(self) -> SampleSet:
        """
        The set of samples :py:class:`genomics_data_index.storage.SampleSet` for which the query does not match.

        :returns: A SampleSet defining the samples where the query does not match.
        """
        pass

    @abc.abstractmethod
    def join(self, data_frame: pd.DataFrame, sample_ids_column: str = None,
             sample_names_column: str = None, default_isa_kind: str = 'names',
             default_isa_column: str = None) -> SamplesQuery:
        """
        Joins the passed dataframe onto the current query using the passed column name in the dataframe.
        The column can either contain sample IDs (in sample_ids_column) or sample names (sample_names_column).
        This will modify the universe set to the subset of samples found within the passed data frame.

        :param data_frame: The data frame to join on.
        :param sample_ids_column: The column name in the data frame containing internal sample IDs used for joining.
        :param sample_names_column: The column name in the data frame containing sample names to join on.
                                    Internally these will be mapped to Sample IDs.
        :param default_isa_kind: The default 'kind' parameter to isa(data, kind=kind, ...) queries. This is used
                                 to simplify queries after joining with a dataframe in cases where you want
                                 isa() to select by a particular column in the dataframe.
        :param default_isa_column: If default_isa_kind is set to 'dataframe' sets the default column to use for
                                   isa() queries.
        :return: A SamplesQuery representing the query joined with the data frame.
        """
        pass

    @abc.abstractmethod
    def toframe(self, include_present: bool = True, include_unknown: bool = False,
                include_absent: bool = False) -> pd.DataFrame:
        """
        Converts the selected set of samples to a DataFrame with one row per sample. By default only samples selected
        by this query will be returned (include_present = True). Setting include_absent to True will include all samples
        where the current query is False, while include_unknown will return all samples where it is unknown if the
        query is True or False (that is, it is unknown if these samples match the query).

        :param include_present: Whether or not samples matching the query should be included.
        :param include_unknown: Whether or not samples where it is unknown if they match the query or not should be included.
        :param include_absent: Whether or not samples absent from this query (but in the universe of samples) should be
        included.

        :return: A DataFrame of the samples in this query, one row per sample. The 'Status' column of the dataframe
                 contains the status (Present, Absent, or Unknown) of the sample.
        """
        pass

    @abc.abstractmethod
    def summary(self) -> pd.DataFrame:
        """
        Summarizes the selected samples in a DataFrame. This includes the number present and absent in the selection
        as well as percents.

        :return: A DataFrame summarizing the selected samples.
        """
        pass

    @abc.abstractmethod
    def select_absent(self) -> SamplesQuery:
        """
        Creates a new query with only the absent samples selected.
        :return: A new query with only the absent samples selected.
        """
        pass

    @abc.abstractmethod
    def select_present(self) -> SamplesQuery:
        """
        Creates a new query with only the present samples selected (minus unknowns).
        :return: A new query with only the present samples selected (minus unknowns).
        """
        pass

    @abc.abstractmethod
    def select_unknown(self) -> SamplesQuery:
        """
        Creates a new query with only the unknown samples selected.
        :return: A new query with only the unknown samples selected.
        """
        pass

    @abc.abstractmethod
    def features_summary(self, kind: str = 'mutations', selection: str = 'all',
                         include_present_features: bool = True, include_unknown_features: bool = False,
                         include_unknown_samples: bool = False, include_unknown_no_present_samples: bool = False,
                         **kwargs) -> pd.DataFrame:
        """
        Summarizes the selected features in a DataFrame. Please specify the kind of features with the kind parameter.

        The `selection` parameter is used to define how to select features in the set of samples. If set to `unique`
        then only those features unique to the selected samples will be returned. Otherwise, if set to `all` then all
        features found in the selected samples are returned (even those that are also found in other samples in the universe).

        For example, say you have a set of samples S = {A, B, C} and features F = {1:A:C, 2:G:T, 3:C:T}. Lets
        also assume that sample A has {1:A:C, 2:G:T}, sample B has {2:G:T, 3:C:T} and sample C has {2:G:T}.
        Also lets assume we have a query q = {A, B} (that is a query has selected two samples A and B).

        Then q.summary_features(selection='all') will return a data frame with features {1:A:C, 2:G:T, 3:C:T} (since
        those are found in either sample A or B).

        However, q.summary_features(selection='unique') will return a data frame with features {1:A:C, 3:C:T} since
        these features are only found in samples A and B (the feature 2:G:T is found in sample C).

        :param kind: The kind of feature to summarize. By default this is *mutations*.
        :param selection: The method used to select features. Use 'all' (default) to select all features present in
                          at least one of the selected samples in this query. Use 'unique' to select only those features
                          that are present in the selected samples of this query
                          (and nowhere else in the universe of samples).
        :param include_present_features: Will determine if present (i.e., not unknown) features should be included.
        :param include_unknown_features: Will determine if unknown features should be included.
        :param include_unknown_samples: Whether or not counts for those samples where it is unknown if they have a
                                        a feature should be included.
        :param include_unknown_no_present_samples: Whether or not counts for features where there are some unknowns but
                                                   no present samples should be included.
        :param **kwargs: Additional keyword arguments. Please see the documentation for the underlying implementation.

        :return: A DataFrame summarizing the features within the selected samples.
        """
        pass

    @abc.abstractmethod
    def features_comparison(self, sample_categories: Union[List[SamplesQuery], List[SampleSet], str],
                            category_prefixes: List[str] = None,
                            categories_kind: str = 'samples',
                            kind: str = 'mutations',
                            unit: str = 'percent',
                            category_samples_threshold: int = None,
                            include_unknown_samples: bool = False, include_unknown_no_present_samples: bool = False,
                            use_only_samples_in_categories: bool = True,
                            **kwargs) -> pd.DataFrame:
        """
        Creates a dataframe which compares different categories of samples with each other with respect to features.

        For example, if kind=='mutations', compare_kind == 'percent' and there are two sample_categories then
        this will return dataframe like:

        | Mutation     | Total | Category1_percent | Category2_percent | Category1_total | Category2_total |
        | ref:100:A:T  | 10    | 50%               | 100%              | 8               | 2               |
        | ref:200:CT:C | 10    | 100%              | 0%                | 8               | 2               |
        | ...          | ...   | ...               | ...               | 8               | 2               |

        Here, "Category1_percent" is the percent of samples in Category1 that have this mutation/feature
        (50% or 4 out of 8 samples in Category1). "Category2_percent" is the percent of samples in Category2 with the
        feature (100% or 2 out of 2 samples in Category2).

        "Category1_total" and "Category2_total" are the total samples in each category. "Total" is the total
        samples in the overall query that form the universe from which we are defining "Category1" and "Category2".

        Additional columns "Category1_Unknown percent" and "Category1_Present and Unknown percent" will be filled in if
        include_unknown_samples=True, which represents the number of samples where this feature is "Unknown/Missing"
        or where the feature is either "Present" or "Unknown" respectively. If include_unknown_samples=False these
        will all be NA.

        Note: since categories are defined based on sample queries, there is no enforcement that categories are
        mutually exclusive (that is, "Category1_total" + "Category2_total" will not always equal "Total"). This
        is done on purpose in case the categories you wish to compare are not mutually exclusive.

        :param sample_categories: The different categories to compare. Either specify as lists of SampleQuery, SampleSet
                                  or a string.
        :param kind: The kind of features to compare.
        :param categories_kind: The kind of category to use ("sample_set", or "dataframe").
        :param category_prefixes: The prefixes to use for the different categories (defaults to 1, 2, 3, ...).
        :param use_only_samples_in_categories: Whether or not to only use samples in categories for comparison
                                               or to use all selected samples. If all selected samples are used
                                               then it's possible to include features in the comparison table
                                               where no samples are found in any category.
        :param unit: The type of data to compare in each category (either 'percent', 'proportion', or 'count').
        :param category_samples_threshold: A threshold on the number of samples in a category for it to be considered.
        :param include_unknown_samples: Whether or not counts for those samples where it is unknown if they have a
                                        a feature should be included.
        :param include_unknown_no_present_samples: Whether or not counts for features where there are some unknowns but
                                                   no present samples should be included.
        :param **kwargs: Additional keyword arguments. Please see the documentation for the underlying implementation.
        :return: A dataframe comparing each category with respect to the differences in features.
        """
        pass

    @abc.abstractmethod
    def tofeaturesset(self, kind: str = 'mutations', selection: str = 'all',
                      include_present_features: bool = True, include_unknown_features: bool = False) -> Set[str]:
        """
        Returns all features as a set of strings for the selected samples.

        :param kind: The kind of feature to summarize. By default this is *mutations*.
        :param selection: The method used to select features. Use 'all' (default) to select all features.
                  Use 'unique' to select only those features unique to the selected samples.
        :param include_present_features: Will determine if present (i.e., not unknown) features should be included.
        :param include_unknown_features: Will determine if unknown features should be included.

        :return: A set of feature identifiers derived from the selected samples.
        """
        pass

    @abc.abstractmethod
    def intersect(self, sample_set: SampleSet, query_message: str = None) -> SamplesQuery:
        """
        Intersects this SamplesQuery with the given :py:class:`genomics_data_index.storage.SampleSet`.
        Appends the given *query_message* to the list of query messages.

        :param sample_set: The :py:class:`genomics_data_index.storage.SampleSet` to intersect.
        :param query_message: A message to include as part of the query messages.

        :return: A new SamplesQuery which is the intersection of this query and the passed set.
        """
        pass

    @abc.abstractmethod
    def and_(self, other: SamplesQuery) -> SamplesQuery:
        """
        Performs an AND operation (intersection) on this query with another query.

        That is if A and B are sample queries then C = A.and_(B) means that C consists of samples in the intersection
        of A and B (C consists of samples that are in A AND B).

        :param other: The other SamplesQuery to use.

        :return: A new SamplesQuery which is the intersection of this query and the other query.
        """
        pass

    @abc.abstractmethod
    def or_(self, other: SamplesQuery) -> SamplesQuery:
        """
        Performs an OR operation (union) on this query with another query.

        That is if A and B are sample queries then C = A.or_(B) means that C consists of samples in the union
        of A and B (C consists of samples that are in A OR B).

        :param other: The other SamplesQuery to use.

        :return: A new SamplesQuery which is the union of this query and the other query.
        """
        pass

    @abc.abstractmethod
    def subsample(self, k: Union[int, float], include_unknown: bool = False, seed: int = None) -> SamplesQuery:
        """
        Select a random subsample of genomic samples from this query.
        :param k: If k >= 1, select this number of samples. If k < 1, select this proportion of samples
                  (e.g., 0.5 means select 50% of the samples in the query).
        :param include_unknown: Whether unknown samples should be included in the subsample. If True,
                                then the value given by n will apply to both unknown and present samples.
        :param seed: The seed for the random number generator (defaults to picking a random seed).
        :return: A new SamplesQuery that selects a random subsample of the genomic samples in the current query.
        """
        pass

    @abc.abstractmethod
    def build_tree(self, kind: str, **kwargs) -> SamplesQuery:
        """
        Builds a tree from the samples contained in this query using the passed *kind* of features.

        :param kind: The *kind* of features to use.
        :param **kwargs: See documentation for the underlying implementation.

        :return: A new SamplesQuery which has a tree attached to it to be used for additional querying.
        """
        pass

    @abc.abstractmethod
    def join_tree(self, tree: Tree, kind='mutation', **kwargs) -> SamplesQuery:
        """
        Joins a tree to this SamplesQuery representing a tree defined by the passed *kind* of features.

        :param tree: The tree to join to this query.
        :param kind: The *kind* of features that were used to build the tree
                      (used to define what type of tree we are joining/the distance units).
        :param **kwargs: See documentation for the underlying implementation.

        :return: A new SamplesQuery which has the passed tree attached to it to be used for additional querying.
        """
        pass

    @abc.abstractmethod
    def reset_universe(self, include_unknown: bool = True) -> SamplesQuery:
        """
        Resets the *universe* set to be the set of currently selected samples.
        That is, if `A` is a SamplesQuery consisting of some selected samples, then
        `B = A.reset_universe()` implies that `B.universe_set == A`.
        :param include_unknown: Whether or not unknown matches should be included in the universe.
        :return: A SamplesQuery with the universe reset to whatever is selected.
        """
        pass

    @abc.abstractmethod
    def set_universe(self, universe_set: SampleSet, query_message: str = None) -> SamplesQuery:
        """
        Sets the *universe* set to be equal to the passed set. Will intersect any selected/unknown sample sets.
        That is, if `A` is a SamplesQuery consisting of some selected samples and `U` is a universe of samples, then
        `B = A.set_universe(U)` implies that: (1) `B.universe_set == U`, (2) `B.sample_set == A.sample_set.intersect(U)`,
        (3) `B.unknown_set == A.unknown_set.intersect(U)`, and (4) `B.absent_set == A.absent_set.intersect(U)`.
        :param universe_set: The new universe set.
        :param query_message: An (optional) message to append to the query message string representing this operation.
        :return: A SamplesQuery with the universe set to the passed set.
        """
        pass

    @abc.abstractmethod
    def hasa(self, property: Union[QueryFeature, str, pd.Series, List[QueryFeature], List[str]],
             kind='mutation') -> SamplesQuery:
        """
        Queries for samples that have a (**hasa**) particular feature/property. That is if `A` is a SamplesQuery
        and `m` is a mutation, then A.hasa(m) selects all those samples of A that have the mutation m.

        :param property: The particular property/feature to query by. This can be either QueryFeature defining
        the particular feature, a string defining the feature, or a pandas Series consisting of boolean values
        which is the result of a DataFrame selection expression. A list of QueryFeature or strings can also be
        passed, which will be interpreted as hasa(feature1) AND hasa(feature2) AND ...
        :param kind: The kind of *property* that was passed.

        :return: A new SamplesQuery consisting of those samples that have the passed property.
        """
        pass

    def has(self, property: Union[QueryFeature, str, pd.Series, List[QueryFeature], List[str]],
             kind='mutation') -> SamplesQuery:
        """
        Queries for samples that have a particular property. Synonym for hasa().
        """
        return self.hasa(property=property, kind=kind)

    @abc.abstractmethod
    def complement(self):
        """
        Returns the complement of the selected set of samples in the SamplesQuery.
        That is if A is a SamplesQuery with U = A.universe_set, then B = A.complement() will result in a
        SamplesQuery, B, consisting of all those samples in U not in A.

        :return: The complement of this SamplesQuery.
        """
        pass

    def within(self, data: Union[str, List[str], SamplesQuery, SampleSet], **kwargs) -> SamplesQuery:
        """
        Queries for samples within a particular distance. This is identical to calling
        isin(data, kind='distance', ...). Used to help make code easier to read.

        :param data: The data to use for selecting samples by (e.g., the samples being used to measure distance against).
        :param **kwargs: Other arguments for the the internal implementations.
        :return: A SamplesQuery with the matched samples.
        """
        return self.isin(data=data, kind='distance', **kwargs)

    @abc.abstractmethod
    def _within_distance(self, data: Union[str, List[str], SamplesQuery, SampleSet], distance: float, units: str,
                         **kwargs) -> SamplesQuery:
        pass

    @abc.abstractmethod
    def _distance_units(self) -> List[str]:
        pass

    @abc.abstractmethod
    def isa(self, data: Union[str, List[str], SamplesQuery, SampleSet], kind: Union[IsaKind, str] = 'sample',
            **kwargs) -> SamplesQuery:
        """
        Queries for samples which are a particular type/belong to a particular category.
        Read as "subset samples which are a (isa) particular type defined by 'data'".
        The default implementation will select samples by sample name but this type of query is also useful when used
        with an attached dataframe (at which point it selects based on matches to a column,
        see documentation for :py:class:`genomics_data_index.api.query.impl.DataFrameSamplesQuery`).

        :param data: The data to match.
        :param kind: The particular kind of data passed.
        :**kwargs: Arguments for the underlying implementation.
        :return: A SamplesQuery with the matched samples.
        """
        pass

    def isan(self, data: Union[str, List[str], SamplesQuery, SampleSet], kind: str = None,
             **kwargs) -> SamplesQuery:
        """
        Synonym for isa()
        """
        return self.isa(data=data, kind=kind, **kwargs)

    @abc.abstractmethod
    def isin(self, data: Union[str, List[str], pd.Series, SamplesQuery, SampleSet], kind: str = None,
             **kwargs) -> SamplesQuery:
        """
        Queries for samples which are in (isin) the passed data. This can have a number of interpretations
        depending on the passed kind
        1. If kind == 'sample' (or 'samples') then isin(['Name']) returns a set of samples with one of the passed names.
        2. If kind == 'distance' (or 'distances') then isin(['Name']) returns a set of samples that are within
            a particular distance of the passed samples.

        :param data: The data to match. This could be a list of strings (representing sample names) or a SampleSet or
                    SamplesQuery representing a set of samples.
        :param kind: The particular kind of data passed.
        :**kwargs: Arguments for the underlying implementation.
        :return: A SamplesQuery with the matched samples.
        """
        pass

    @abc.abstractmethod
    def _get_sample_names_query_infix_from_data(self, data: Union[str, List[str], pd.Series, SamplesQuery, SampleSet]
                                                ) -> Tuple[Set[str], str]:
        pass

    @abc.abstractmethod
    def _can_handle_isin_kind(self, kind: str) -> bool:
        pass

    @abc.abstractmethod
    def _can_handle_isa_kind(self, kind: str) -> bool:
        pass

    @abc.abstractmethod
    def _isa_kinds(self) -> List[str]:
        pass

    @abc.abstractmethod
    def _isin_kinds(self) -> List[str]:
        pass

    @abc.abstractmethod
    def to_distances(self, kind: str, **kwargs) -> Tuple[np.ndarray, List[str]]:
        """
        Returns a distance-matrix representing the pairwise distances between all selected samples.

        :param kind: The features kind used to generate the distance matrix (e.g,. 'mutations', or 'kmer').
        :param **kwargs: Additional arguments depending on the implementation.

        :return: A Tuple consisting of a pairwise distance matrix (as a NumPy array) and a list of labels of the distance
                 matrix (sample names).
        """
        pass

    @abc.abstractmethod
    def _get_has_kinds(self) -> List[str]:
        pass

    @abc.abstractmethod
    def is_empty(self, include_unknown=False) -> bool:
        """
        Whether or not the selected set of samples is empty.
        :param include_unknown: Whether or not to include samples with unknown status.
        :return: True if the selected set of samples is empty, False otherwise.
        """
        pass

    @abc.abstractmethod
    def query_expression(self) -> str:
        """
        A string representing the series of queries performed to select the given samples in this SamplesQuery.
        For example if A is a SamplesQuery then `A.query_expression() == "hasa('mutation') AND hasa('mutation2')"
        would mean that the SamplesQuery A represents samples that have both 'mutation' and 'mutation2'.

        :return: A string representing the set of queries performed to generate this set.
        """
        pass

    @property
    @abc.abstractmethod
    def tree(self):
        """
        If this SamplesQuery has a joined tree, then returns the tree.

        :return: An ete3.Tree that has been joined to this SamplesQuery (or raises an exception if no tree).
        """
        pass

    @abc.abstractmethod
    def has_tree(self) -> bool:
        """
        Whether or not this SamplesQuery has a joined tree.

        :return: True if this SamplesQuery has a joined tree, False otherwise.
        """
        pass

    @abc.abstractmethod
    def tolist(self, names: bool = True, include_present: bool = True,
               include_unknown: bool = False, include_absent: bool = False) -> Union[List[str], List[int]]:
        """
        Converts the set of selected samples into a list of either sample names or sample IDs.

        :param names: If True (default) return a list of sample names as strings, if False return a list of sample IDs.
        :param include_present: If True (default) include selected samples.
        :param include_unknown: If True, include unknown samples.
        :param include_absent: If True, include absent samples from selection.
        :return: A list of sample names or IDs.
        """
        pass

    @abc.abstractmethod
    def toset(self, names: bool = True, include_present: bool = True,
              include_unknown: bool = False, include_absent: bool = False) -> Union[Set[str], Set[int]]:
        """
        Converts the set of selected samples into a set of either sample names or sample IDs.

        :param names: If True (default) return a set of sample names as strings, if False return a set of sample IDs.
        :param include_present: If True (default) include selected samples.
        :param include_unknown: If True, include unknown samples.
        :param include_absent: If True, include absent samples from selection.
        :return: A set of sample names or IDs.
        """
        pass

    def __invert__(self):
        """
        Performs an **~** (invert, complement) operation on a SamplesQuery object.
        If A is a SamplesQuery object then `~A` is equivalent to `A.complement()`.

        :return: The complement of a SamplesQuery object.
        """
        return self.complement()

    def __and__(self, other):
        """
        Performs an **&** (and) operation (intersection) between two different SamplesQuery objects.
        If A and B are two SamplesQuery objects then `A & B` is equivalent to `A.and_(B)`.

        :return: The and (intersection) of two SamplesQuery objects.
        """
        return self.and_(other)

    def __or__(self, other):
        """
        Performs an **|** (or) operation (union) between two different SamplesQuery objects.
        If A and B are two SamplesQuery objects then `A | B` is equivalent to `A.or_(B)`.

        :return: The or (union) of two SamplesQuery objects.
        """
        return self.or_(other)

    @abc.abstractmethod
    def __len__(self):
        """
        The number of selected samples.

        :return: The number of selected samples.
        """
        pass



from __future__ import annotations

import logging
import random
from typing import Union, List, Set, Tuple, Dict

import numpy as np
import pandas as pd
from ete3 import Tree

from genomics_data_index.api.query.SamplesQuery import SamplesQuery
from genomics_data_index.api.query.features.FeaturesFromIndexComparator import FeaturesFromIndexComparator
from genomics_data_index.api.query.features.MLSTFeaturesComparator import MLSTFeaturesComparator
from genomics_data_index.api.query.features.MutationFeaturesFromIndexComparator import \
    MutationFeaturesFromIndexComparator
from genomics_data_index.api.query.impl.DataFrameSamplesQuery import DataFrameSamplesQuery
from genomics_data_index.api.query.impl.QueriesCollection import QueriesCollection
from genomics_data_index.api.query.impl.TreeSamplesQueryFactory import TreeSamplesQueryFactory
from genomics_data_index.api.query.kind.IsaKind import IsaKind
from genomics_data_index.api.query.kind.isa.DelegateIsaKind import DelegateIsaKind
from genomics_data_index.configuration.connector import DataIndexConnection
from genomics_data_index.storage.SampleSet import SampleSet
from genomics_data_index.storage.model.QueryFeature import QueryFeature
from genomics_data_index.storage.model.QueryFeatureFactory import QueryFeatureFactory
from genomics_data_index.storage.service.KmerService import KmerService

logger = logging.getLogger(__name__)


class SamplesQueryIndex(SamplesQuery):
    """
    The main class implementing :py:class:`genomics_data_index.api.query.SamplesQuery`. This is used
    to store selections of sets of samples as well as operations to select subsets of samples.
    """

    HAS_KINDS = ['mutation', 'mutations', 'mlst']
    SUMMARY_FEATURES_KINDS = ['mutations', 'mlst']
    FEATURES_SELECTIONS = ['all', 'unique']
    ISIN_TYPES = ['sample', 'samples', 'distance', 'distances']
    ISA_TYPES = ['sample', 'samples']
    DISTANCES_UNITS = ['kmer_jaccard']
    CATEGORY_KINDS = ['samples']

    def __init__(self, connection: DataIndexConnection,
                 universe_set: SampleSet,
                 sample_set: SampleSet,
                 unknown_set: SampleSet,
                 queries_collection: QueriesCollection = QueriesCollection.create_empty()):
        """
        Builds a new SamplesQueryIndex from the given information. In most normal operations SamplesQuery objects
        are not created directly but are instead created from an :py:class:`genomics_data_index.api.GenomicsDataIndex`
        object or from operations applied to a SamplesQuery.

        :param connection: A connection to a database containing samples.
        :param universe_set: The :py:class:`genomics_data_index.storage.SampleSet` representing a set of samples defining
                       the universe (used for e.g., complement() operations).
        :param sample_set: The :py:class:`genomics_data_index.storage.SampleSet` representing the set of selected samples.
        :param unknown_set: The :py:class:`genomics_data_index.storage.SampleSet` representing the set of unknown samples.
        :param queries_collection: A collection of strings representing the queries performed to arrive at this SamplesQuery.
        :return: A new SamplesQueryIndex object.
        """
        super().__init__()
        self._query_connection = connection
        self._universe_set = universe_set
        self._sample_set = sample_set
        self._unknown_set = unknown_set
        self._queries_collection = queries_collection
        self._query_feature_factory = QueryFeatureFactory.instance()

    @property
    def universe_set(self) -> SampleSet:
        return self._universe_set

    @property
    def sample_set(self) -> SampleSet:
        return self._sample_set

    def select_absent(self) -> SamplesQuery:
        queries_collection = self._queries_collection.append('select_absent')
        return self._create_from(sample_set=self.absent_set, universe_set=self.universe_set,
                                 unknown_set=SampleSet.create_empty(),
                                 queries_collection=queries_collection)

    def select_present(self) -> SamplesQuery:
        queries_collection = self._queries_collection.append('select_present')
        return self._create_from(sample_set=self.sample_set, universe_set=self.universe_set,
                                 unknown_set=SampleSet.create_empty(),
                                 queries_collection=queries_collection)

    def select_unknown(self) -> SamplesQuery:
        queries_collection = self._queries_collection.append('select_unknown')
        return self._create_from(sample_set=self.unknown_set, universe_set=self.universe_set,
                                 unknown_set=SampleSet.create_empty(),
                                 queries_collection=queries_collection)

    @property
    def unknown_set(self) -> SampleSet:
        return self._unknown_set

    @property
    def _absent_set(self) -> SampleSet:
        return self._universe_set.minus([self._sample_set, self._unknown_set])

    @property
    def absent_set(self) -> SampleSet:
        return self._absent_set

    def reset_universe(self, include_unknown: bool = True) -> SamplesQuery:
        if include_unknown:
            universe_set = self._sample_set.union(self._unknown_set)
            unknown_set = self._unknown_set
        else:
            universe_set = self._sample_set
            unknown_set = SampleSet.create_empty()

        return self._create_from(sample_set=self._sample_set, universe_set=universe_set,
                                 unknown_set=unknown_set,
                                 queries_collection=self._queries_collection)

    def set_universe(self, universe_set: SampleSet, query_message: str = None) -> SamplesQuery:
        if universe_set is None:
            universe_set = SampleSet.create_empty()

        if query_message is None:
            query_message = f'set_universe({len(universe_set)} samples)'

        queries_collection = self._queries_collection.append(query_message)

        sample_set = self.sample_set.intersection(universe_set)
        unknown_set = self.unknown_set.intersection(universe_set)
        return self._create_from(sample_set=sample_set, universe_set=universe_set,
                                 unknown_set=unknown_set, queries_collection=queries_collection)

    def join(self, data_frame: pd.DataFrame, sample_ids_column: str = None,
             sample_names_column: str = None, default_isa_kind: str = 'sample',
             default_isa_column: str = None) -> SamplesQuery:
        if sample_ids_column is None and sample_names_column is None:
            raise Exception('At least one of sample_ids_column or sample_names_column must be set.')
        elif sample_ids_column is not None:
            return DataFrameSamplesQuery.create_with_sample_ids_column(sample_ids_column=sample_ids_column,
                                                                       data_frame=data_frame,
                                                                       wrapped_query=self,
                                                                       connection=self._query_connection,
                                                                       default_isa_kind=default_isa_kind,
                                                                       default_isa_column=default_isa_column)
        else:
            return DataFrameSamplesQuery.create_with_sample_names_column(sample_names_column=sample_names_column,
                                                                         data_frame=data_frame,
                                                                         wrapped_query=self,
                                                                         connection=self._query_connection,
                                                                         default_isa_kind=default_isa_kind,
                                                                         default_isa_column=default_isa_column)

    def join_tree(self, tree: Tree, kind='mutation', **kwargs) -> SamplesQuery:
        return TreeSamplesQueryFactory.instance().join_tree(tree=tree,
                                                            kind=kind,
                                                            database_connection=self._query_connection,
                                                            wrapped_query=self,
                                                            **kwargs)

    def intersect(self, sample_set: SampleSet, query_message: str = None) -> SamplesQuery:
        intersected_set = self._intersect_sample_set(sample_set)
        unknown_set = self._intersect_unknown_set(sample_set)

        if query_message is None:
            query_message = f'intersect(samples={len(sample_set)}'

        queries_collection = self._queries_collection.append(query_message)
        return self._create_from(sample_set=intersected_set, universe_set=self._universe_set,
                                 unknown_set=unknown_set,
                                 queries_collection=queries_collection)

    def _found_in_self_and(self, found_in_other: SampleSet) -> SampleSet:
        return self._sample_set.intersection(found_in_other)

    def _found_in_self_or(self, found_in_other: SampleSet) -> SampleSet:
        return self._sample_set.union(found_in_other)

    def _unknown_in_self_or(self, unknown_in_other: SampleSet, found_either: SampleSet):
        return self._unknown_set.union(unknown_in_other).minus(found_either)

    def _unknown_in_self_and(self, found_in_other: SampleSet, unknown_in_other: SampleSet) -> SampleSet:
        """
        Given the above SampleSets, this returns the set of those unknown in self and the other sets
        (representing another query). This is defined using the help of Kleene's three-valued logic truth tables
        <https://en.wikipedia.org/wiki/Three-valued_logic#Kleene_and_Priest_logics>. Specifically, for two queries A
        and B with sets of samples consisting of either found (True), Unknown, or absent (False), this will return the
        sets of samples in the unknown state (U) in the truth table for AND(A,B).
        :param found_in_other: The SampleSet found in the other query.
        :param unknown_in_other: The SampleSet of unknowns in the other query.
        :return: The SampleSet of unknowns in self AND the other query.
        """
        # Each corresponds to one of three possible combinations in the three-valued truth-table for A and B
        # For those samples that should be in the unknown state (U) for "A AND B".
        unknown_and_unknown = self._unknown_set.intersection(unknown_in_other)
        unknown_and_found = self._unknown_set.intersection(found_in_other)
        found_and_unknown = self._sample_set.intersection(unknown_in_other)
        return unknown_and_unknown.union(unknown_and_found).union(found_and_unknown)

    def _intersect_sample_set(self, other: SampleSet) -> SampleSet:
        return self.sample_set.intersection(other)

    def _union_sample_set(self, other: SampleSet) -> SampleSet:
        return self.sample_set.union(other)

    def _union_unknown_set(self, other: SampleSet) -> SampleSet:
        return self.unknown_set.union(other)

    def _intersect_unknown_set(self, other: SampleSet) -> SampleSet:
        return self.unknown_set.intersection(other)

    def _get_has_kinds(self) -> List[str]:
        return self.HAS_KINDS

    def query_expression(self) -> str:
        return self._queries_collection.query_expression()

    def build_tree(self, kind: str, **kwargs) -> SamplesQuery:
        return TreeSamplesQueryFactory.instance().build_tree(kind=kind,
                                                             database_connection=self._query_connection,
                                                             wrapped_query=self, **kwargs)

    def toframe(self, include_present: bool = True, include_unknown: bool = False,
                include_absent: bool = False) -> pd.DataFrame:
        sample_service = self._query_connection.sample_service
        queries_expression = self._queries_collection.query_expression()

        present_set = self.sample_set if include_present else SampleSet.create_empty()
        absent_set = self.absent_set if include_absent else SampleSet.create_empty()
        unknown_set = self.unknown_set if include_unknown else SampleSet.create_empty()

        return sample_service.create_dataframe_from_sample_set(present_set=present_set,
                                                               absent_set=absent_set,
                                                               unknown_set=unknown_set,
                                                               queries_expression=queries_expression)

    def summary(self) -> pd.DataFrame:
        present = len(self)
        total = len(self.universe_set)
        unknown = len(self.unknown_set)
        absent = total - present - unknown
        per_present = (present / total) * 100
        per_absent = (absent / total) * 100
        per_unknown = (unknown / total) * 100

        return pd.DataFrame([{
            'Query': self._queries_collection.query_expression(),
            'Present': present,
            'Absent': absent,
            'Unknown': unknown,
            'Total': total,
            '% Present': per_present,
            '% Absent': per_absent,
            '% Unknown': per_unknown,
        }])

    def features_summary(self, kind: str = 'mutations', selection: str = 'all',
                         include_present_features: bool = True, include_unknown_features: bool = False,
                         include_unknown_samples: bool = False, include_unknown_no_present_samples: bool = False,
                         **kwargs) -> pd.DataFrame:
        if kind == 'mutations':
            features_summarizier = MutationFeaturesFromIndexComparator(connection=self._query_connection,
                                                                       include_unknown=include_unknown_features,
                                                                       include_present=include_present_features,
                                                                       include_unknown_samples=include_unknown_samples,
                                                                       include_unknown_no_present_samples=include_unknown_no_present_samples,
                                                                       **kwargs)
        elif kind == 'mlst':
            features_summarizier = MLSTFeaturesComparator(connection=self._query_connection,
                                                          include_unknown=include_unknown_features,
                                                          include_present=include_present_features,
                                                          include_unknown_samples=include_unknown_samples,
                                                          include_unknown_no_present_samples=include_unknown_no_present_samples,
                                                          **kwargs)
        else:
            raise Exception(f'Unsupported value kind=[{kind}]. Must be one of {self.SUMMARY_FEATURES_KINDS}.')

        if selection == 'all':
            return features_summarizier.summary(self.sample_set)
        elif selection == 'unique':
            return features_summarizier.unique_summary(self.sample_set,
                                                       other_set=self.universe_set.minus(self.sample_set))
        else:
            raise Exception(f'selection=[{selection}] is unknown. Must be one of {self.FEATURES_SELECTIONS}')

    def features_comparison(self, sample_categories: Union[List[SamplesQuery], List[SampleSet], str],
                            category_prefixes: List[str] = None,
                            categories_kind: str = 'samples',
                            kind: str = 'mutations',
                            unit: str = 'percent',
                            category_samples_threshold: int = None,
                            include_unknown_samples: bool = False, include_unknown_no_present_samples: bool = False,
                            use_only_samples_in_categories: bool = True,
                            **kwargs) -> pd.DataFrame:
        if kind == 'mutations':
            features_comparator = MutationFeaturesFromIndexComparator(connection=self._query_connection,
                                                                      include_unknown=False,
                                                                      include_present=True,
                                                                      include_unknown_samples=include_unknown_samples,
                                                                      include_unknown_no_present_samples=include_unknown_no_present_samples,
                                                                      **kwargs)
        elif kind == 'mlst':
            features_comparator = MLSTFeaturesComparator(connection=self._query_connection,
                                                         include_unknown=False,
                                                         include_present=True,
                                                         include_unknown_samples=include_unknown_samples,
                                                         include_unknown_no_present_samples=include_unknown_no_present_samples,
                                                         **kwargs)
        else:
            raise Exception(f'Unsupported value kind=[{kind}]. Must be one of {self.SUMMARY_FEATURES_KINDS}.')

        if categories_kind not in self.CATEGORY_KINDS:
            raise Exception(f'Unknown categories_kind={categories_kind}. Must be one of {self.CATEGORY_KINDS}. '
                            f'Are you sure you have the correct instance of the query class: {self.__class__}?')

        if not isinstance(sample_categories, list):
            raise Exception(
                f'sample_categories={sample_categories} must be a list (to iterate in a well-defined order)')
        else:
            categories = []
            for sample_category in sample_categories:
                if isinstance(sample_category, SamplesQuery):
                    sample_set = sample_category.sample_set
                elif isinstance(sample_category, SampleSet):
                    sample_set = sample_category
                else:
                    raise Exception(f'Unknown type={type(sample_category)} in sample_cateogires={sample_categories}. '
                                    f'Type must be either {SamplesQuery.__name__} or {SampleSet.__name__}')

                categories.append(sample_set)

            return features_comparator.features_comparison(selected_samples=self.sample_set,
                                                           sample_categories=categories,
                                                           category_prefixes=category_prefixes,
                                                           category_samples_threshold=category_samples_threshold,
                                                           use_only_samples_in_categories=use_only_samples_in_categories,
                                                           unit=unit)

    def tofeaturesset(self, kind: str = 'mutations', selection: str = 'all',
                      include_present_features: bool = True, include_unknown_features: bool = False) -> Set[str]:
        return set(self.features_summary(kind=kind, selection=selection, ignore_annotations=True,
                                         include_present_features=include_present_features,
                                         include_unknown_features=include_unknown_features).index)

    def and_(self, other):
        if isinstance(other, SamplesQuery):
            present_set = self._found_in_self_and(other.sample_set)
            unknown_set = self._unknown_in_self_and(other.sample_set, other.unknown_set)
            universe_set = self.universe_set.union(other.universe_set)
            queries_collection = self._queries_collection.append(str(other))
            return self._create_from(present_set, universe_set=universe_set,
                                     unknown_set=unknown_set,
                                     queries_collection=queries_collection)
        else:
            raise Exception(f'Cannot perform an "and" on object {other}')

    def or_(self, other: SamplesQuery) -> SamplesQuery:
        if isinstance(other, SamplesQuery):
            present_set = self._found_in_self_or(other.sample_set)
            unknown_set = self._unknown_in_self_or(unknown_in_other=other.unknown_set, found_either=present_set)
            universe_set = self.universe_set.union(other.universe_set)

            queries_collection = self._queries_collection.append(f'OR({str(other)}')
            return self._create_from(present_set, universe_set=universe_set,
                                     unknown_set=unknown_set,
                                     queries_collection=queries_collection)
        else:
            raise Exception(f'Cannot perform an "or" on object {other}')

    def subsample(self, k: Union[int, float], include_unknown: bool = False, seed: int = None) -> SamplesQuery:
        if k is None:
            raise Exception(f'k={k} must not be None')
        elif k < 0:
            raise Exception(f'k={k} must not be negative')

        # I set a random seed here using random.randint so that I can display the seed used
        # As part of the query string text. That way, someone could, later on, reproduce the same
        # subsample even if they didn't explicity set the seed.
        if seed is None:
            seed = random.randint(0, 10 ** 10)

        initial_sample_set = self.sample_set
        if include_unknown:
            initial_sample_set = initial_sample_set.union(self.unknown_set)

        if k < 1:
            k = round(k * len(initial_sample_set))
        elif k > len(initial_sample_set):
            raise Exception(f'k={k} is greater than number of samples in query={len(initial_sample_set)}')
        else:
            k = round(k)

        # There might be a more efficient way to select a random subsample directly from a roaring bitmap
        # instead of converting to a list and back, but I wasn't able to find one.
        initial_samples_list = list(initial_sample_set)
        random.seed(seed)
        subsamples_list = random.sample(initial_samples_list, k=k)
        subsamples = SampleSet(subsamples_list)

        return self.intersect(subsamples, f'subsample(k={k}, seed={seed})')

    def is_empty(self, include_unknown=False) -> bool:
        if include_unknown:
            return self.sample_set.is_empty() and self.unknown_set.is_empty()
        else:
            return self.sample_set.is_empty()

    def complement(self):
        complement_present_set = self.absent_set
        query_collection = self._queries_collection.append('complement')
        return self._create_from(sample_set=complement_present_set, universe_set=self._universe_set,
                                 unknown_set=self._unknown_set,
                                 queries_collection=query_collection)

    @property
    def tree(self):
        raise Exception(f'No tree exists for {self.__class__}.'
                        f' Perhaps you should try to run build_tree() first to build a tree.')

    def has_tree(self) -> bool:
        return False

    def __len__(self):
        return len(self.sample_set)

    def __str__(self) -> str:
        universe_length = len(self.universe_set)
        unknown_length = len(self.unknown_set)
        selected_length = len(self)
        if universe_length > 0:
            percent_selected = f'{(selected_length / universe_length) * 100:0.0f}'
            percent_unknown = f'{(unknown_length / universe_length) * 100:0.0f}'
        else:
            percent_selected = 'NA'
            percent_unknown = 'NA'

        return f'<{self.__class__.__name__}[selected={percent_selected}% ' \
               f'({selected_length}/{universe_length}) samples, unknown={percent_unknown}% ' \
               f'({unknown_length}/{universe_length}) samples]>'

    def tolist(self, names: bool = True, include_present: bool = True,
               include_unknown: bool = False, include_absent: bool = False) -> Union[List[str], List[int]]:
        sample_set = SampleSet.create_empty()
        if include_present:
            sample_set = sample_set.union(self.sample_set)
        if include_absent:
            sample_set = sample_set.union(self.absent_set)
        if include_unknown:
            sample_set = sample_set.union(self.unknown_set)

        if names:
            sample_service = self._query_connection.sample_service
            return [s.name for s in sample_service.find_samples_by_ids(sample_set)]
        else:
            return list(sample_set)

    def toset(self, names: bool = True, include_present: bool = True,
              include_unknown: bool = False, include_absent: bool = False) -> Union[Set[str], Set[int]]:
        return set(self.tolist(names=names,
                               include_present=include_present,
                               include_unknown=include_unknown,
                               include_absent=include_absent))

    def __repr__(self) -> str:
        return str(self)

    def _get_samples_or_empty(self, feature: QueryFeature, feature_id_set_dict: Dict[str, SampleSet]) -> SampleSet:
        if feature.id in feature_id_set_dict:
            return feature_id_set_dict[feature.id]
        else:
            return SampleSet.create_empty()

    def hasa(self, property: Union[QueryFeature, str, pd.Series, List[QueryFeature], List[str]],
             kind='mutation') -> SamplesQuery:
        if isinstance(property, list):
            list_query = self
            for list_property in property:
                list_query = list_query.hasa(list_property, kind=kind)
            return list_query
        elif isinstance(property, QueryFeature):
            query_feature = property
        elif isinstance(property, pd.Series):
            raise Exception(f'The query type {self.__class__.__name__} cannot support querying with respect to a '
                            f'dataframe. Perhaps you could try attaching a dataframe with join() first before querying.')
        elif kind is None:
            raise Exception(f'property=[{property}] is not of type QueryFeature so must set "kind" parameter')
        elif kind == 'mutation' or kind == 'mutations' or kind == 'mlst':
            query_feature = self._query_feature_factory.create_feature(property)
        else:
            raise Exception(f'kind={kind} is not recognized for {self}. Must be one of {self._get_has_kinds()}')

        found_hasa_set_dict = self._query_connection.sample_service.find_sample_sets_by_features([query_feature])
        unknown_hasa_set_dict = self._query_connection.sample_service.find_unknown_sample_sets_by_features(
            [query_feature])

        found_hasa_set = self._get_samples_or_empty(query_feature, found_hasa_set_dict)
        unknown_hasa_set = self._get_samples_or_empty(query_feature, unknown_hasa_set_dict)

        found_in_query = self._found_in_self_and(found_hasa_set)
        unknown_in_query = self._unknown_in_self_and(found_in_other=found_hasa_set, unknown_in_other=unknown_hasa_set)

        # Universe remains the same in this case.
        universe_in_query = self._universe_set

        queries_collection = self._queries_collection.append(query_feature)
        return self._create_from(found_in_query, universe_set=universe_in_query,
                                 unknown_set=unknown_in_query,
                                 queries_collection=queries_collection)

    def _prepare_isin_query_message(self, query_message_prefix: str,
                                    query_msg_infix: str,
                                    additional_messages: str) -> str:
        return f"{query_message_prefix}({query_msg_infix}{additional_messages})"

    def _isin_samples(self, data: Union[str, List[str], pd.Series, SamplesQuery, SampleSet],
                      query_message_prefix: str) -> SamplesQuery:
        present_set, unknown_set, query_infix = self._get_present_unknown_sets_query_infix_from_data(data)
        query_message = self._prepare_isin_query_message(query_message_prefix=query_message_prefix,
                                                         query_msg_infix=query_infix,
                                                         additional_messages='')
        queries_collection = self._queries_collection.append(query_message)
        return self._create_from(sample_set=present_set, universe_set=self._universe_set,
                                 unknown_set=unknown_set,
                                 queries_collection=queries_collection)

    def _within_kmer_jaccard(self, data: Union[str, List[str], pd.Series, SamplesQuery, SampleSet],
                             distance: float, kmer_size: int = 31) -> SamplesQuery:
        additional_messages = f', dist={distance}, k={kmer_size}'
        sample_names, query_infix = self._get_sample_names_query_infix_from_data(data)
        query_message = self._prepare_isin_query_message(query_message_prefix='isin_kmer_jaccard',
                                                         query_msg_infix=query_infix,
                                                         additional_messages=additional_messages)

        kmer_service: KmerService = self._query_connection.kmer_service

        if len(sample_names) == 0:
            sample_set_matches = SampleSet.create_empty()
            unknown_matches = SampleSet.create_empty()
        else:
            sample_set_matches = kmer_service.find_matches_within(sample_names=list(sample_names),
                                                                  distance_threshold=distance,
                                                                  kmer_size=kmer_size,
                                                                  samples_universe=self._sample_set)
            unknown_matches = kmer_service.find_matches_within(sample_names=list(sample_names),
                                                               distance_threshold=distance,
                                                               kmer_size=kmer_size,
                                                               samples_universe=self._unknown_set)
        queries_collection = self._queries_collection.append(query_message)
        return self._create_from(sample_set=sample_set_matches, universe_set=self._universe_set,
                                 unknown_set=unknown_matches,
                                 queries_collection=queries_collection)

    def _within_distance(self, data: Union[str, List[str], pd.Series, SamplesQuery, SampleSet], distance: float,
                         units: str = 'kmer_jaccard', **kwargs) -> SamplesQuery:
        if units == 'kmer_jaccard':
            return self._within_kmer_jaccard(data=data, distance=distance, **kwargs)
        else:
            raise Exception(f'units=[{units}] is not supported. Must be one of {self._distance_units()}. '
                            f'For additional distance queries you perhaps need to build or attach a tree to '
                            f'the query.')

    def _get_present_unknown_sets_query_infix_from_data(self,
                                                        data: Union[str, List[str], pd.Series, SamplesQuery, SampleSet]
                                                        ) -> Tuple[SampleSet, SampleSet, str]:
        if isinstance(data, str) or isinstance(data, list):
            logger.debug(f'data=[{data}] contains sample names')
            if isinstance(data, str):
                query_msg = f"'{data}'"
                data = [data]
            else:
                query_msg = f'{data}'
            samples = self._query_connection.sample_service.get_existing_samples_by_names(data)
            sample_ids = {s.id for s in samples}
            present_set = SampleSet(sample_ids)
        elif isinstance(data, SamplesQuery) or isinstance(data, SampleSet):
            logger.debug(f'data=[{data}] contains sample ids')
            if isinstance(data, SamplesQuery):
                present_set = data.sample_set
                query_msg = f'{data}'
            else:
                present_set = data
                query_msg = f'set({len(data)} samples)'
        else:
            raise Exception(f'Unknown type for data=[{data}. Got type [{type(data)}]. '
                            'Must a string or list of strings (representing sample names) '
                            'or a set of sample IDs (as a SamplesQuery or SampleSet).')
        unknown_set = self.unknown_set.intersection(present_set)
        present_set = self.sample_set.intersection(present_set)
        return present_set, unknown_set, query_msg

    def _get_sample_names_query_infix_from_data(self, data: Union[str, List[str], pd.Series, SamplesQuery, SampleSet]
                                                ) -> Tuple[Set[str], str]:
        if isinstance(data, str) or isinstance(data, list):
            logger.debug(f'data={data} contains sample names')
            if isinstance(data, str):
                query_msg = f"'{data}'"
                data = [data]
            else:
                query_msg = f'{data}'
            sample_names = data
        elif isinstance(data, SamplesQuery) or isinstance(data, SampleSet):
            logger.debug(f'data=[{data}] contains sample ids')
            if isinstance(data, SamplesQuery):
                sample_set = data.sample_set
                query_msg = f'{data}'
            else:
                sample_set = data
                query_msg = f'set({len(data)} samples)'

            samples = self._query_connection.sample_service.find_samples_by_ids(sample_set)
            sample_names = {s.name for s in samples}
        else:
            raise Exception(f'Unknown type for data=[{data}. Got type [{type(data)}]. '
                            'Must a string or list of strings (representing sample names) '
                            'or a set of sample IDs (as a SamplesQuery or SampleSet).')
        return sample_names, query_msg

    def isin(self, data: Union[str, List[str], pd.Series, SamplesQuery, SampleSet], kind: str = 'samples',
             **kwargs) -> SamplesQuery:
        if kind == 'sample' or kind == 'samples':
            return self._isin_samples(data=data, query_message_prefix='isin_samples')
        elif kind == 'distance' or kind == 'distances':
            return self._within_distance(data=data, **kwargs)
        else:
            raise Exception(f'kind=[{kind}] is not supported. Must be one of {self.ISIN_TYPES}')

    def _isin_kinds(self) -> List[str]:
        return self.ISIN_TYPES

    def _isa_kinds(self) -> List[str]:
        return self.ISA_TYPES

    def _can_handle_isa_kind(self, kind: str) -> bool:
        return kind in self.ISA_TYPES

    def _can_handle_isin_kind(self, kind: str) -> bool:
        return kind in self.ISIN_TYPES

    def _distance_units(self) -> List[str]:
        return self.DISTANCES_UNITS

    def _can_handle_distance_units(self, units: str) -> bool:
        return units in self.DISTANCES_UNITS

    def isa(self, data: Union[str, List[str], SamplesQuery, SampleSet], kind: Union[IsaKind, str] = 'sample',
            **kwargs) -> SamplesQuery:
        if kind == 'sample' or kind == 'samples':
            return self._isin_samples(data=data, query_message_prefix='isa_sample')
        elif isinstance(kind, DelegateIsaKind):
            return kind.isa(data, self)
        else:
            raise Exception(f'kind=[{kind}] is not supported. Must be one of {self._isa_kinds()}')

    def to_distances(self, kind: str = 'kmer', **kwargs) -> Tuple[np.ndarray, List[str]]:
        if kind == 'kmer':
            return self._to_distances_kmer(**kwargs)
        else:
            raise Exception(f'kind=[{kind}] is not supported. Must be one of {self.DISTANCES_UNITS}')

    def _to_distances_kmer(self, kmer_size: int = 31, ncores: int = 1) -> Tuple[np.ndarray, List[str]]:
        return self._query_connection.kmer_service.get_distance_matrix(sample_ids=self._sample_set,
                                                                       kmer_size=kmer_size,
                                                                       ncores=ncores)

    def _create_from(self, sample_set: SampleSet, universe_set: SampleSet, unknown_set: SampleSet,
                     queries_collection: QueriesCollection) -> SamplesQuery:

        last_query = queries_collection.last
        extra_msg = f' following the query "{last_query}'
        sample_set = FeaturesFromIndexComparator.remove_unknowns_from_present_samples(
            sample_set=sample_set,
            unknown_set=unknown_set,
            sample_service=self._query_connection.sample_service,
            extra_message=extra_msg)

        return SamplesQueryIndex(connection=self._query_connection,
                                 universe_set=universe_set,
                                 sample_set=sample_set,
                                 unknown_set=unknown_set,
                                 queries_collection=queries_collection)



from pathlib import Path

from genomics_data_index.storage.MaskedGenomicRegions import MaskedGenomicRegions
from genomics_data_index.storage.io.mutation.SequenceFile import SequenceFile

# Generates a mask bed file from a snippy output directory


snippy_dir = Path(snakemake.input.snippy_dir)
output_mask = Path(snakemake.output.mask)

consensus_file = snippy_dir / 'snps.aligned.fa'
vcf_file = snippy_dir / 'snps.vcf.gz'

name, consensus_records = SequenceFile(consensus_file).parse_sequence_file()
mask = MaskedGenomicRegions.from_sequences(consensus_records)
vcf_regions = MaskedGenomicRegions.from_vcf_file(vcf_file)
mask = mask.subtract(vcf_regions)

mask.write(output_mask)



import logging
import shutil
from pathlib import Path
from typing import Tuple, Optional

import pandas as pd

from genomics_data_index.storage.MaskedGenomicRegions import MaskedGenomicRegions
from genomics_data_index.storage.io.SampleData import SampleData
from genomics_data_index.storage.io.mutation.VariationFile import VariationFile
from genomics_data_index.storage.io.mutation.VcfSnpEffAnnotationParser import VcfSnpEffAnnotationParser
from genomics_data_index.storage.model import NUCLEOTIDE_UNKNOWN_TYPE
from genomics_data_index.storage.util import TRACE_LEVEL

logger = logging.getLogger(__name__)


class NucleotideSampleData(SampleData):
    VCF_FRAME_COLUMNS = ['SAMPLE', 'CHROM', 'POS', 'REF', 'ALT', 'TYPE', 'FILE', 'VARIANT_ID']

    def __init__(self, sample_name: str, vcf_file: Optional[Path], vcf_file_index: Optional[Path],
                 mask_bed_file: Optional[Path],
                 preprocessed: bool):
        super().__init__(sample_name=sample_name)
        self._vcf_file = vcf_file
        self._vcf_file_index = vcf_file_index
        self._preprocessed = preprocessed
        self._mask_bed_file = mask_bed_file
        self._snpeff_parser = VcfSnpEffAnnotationParser()

    def _preprocess_mask(self, output_dir: Path) -> Path:
        if self._mask_bed_file is None:
            mask_file = output_dir / f'{self.sample_name_persistence}.bed.gz'
            self._assert_file_not_exists(mask_file, 'Cannot preprocess data')
            mask = MaskedGenomicRegions.empty_mask()
            mask.write(mask_file)
            self._mask_bed_file = mask_file
        return self._mask_bed_file

    def _preprocess_vcf(self, output_dir: Path) -> Tuple[Path, Path]:
        new_file = output_dir / f'{self.sample_name_persistence}.vcf.gz'
        self._assert_file_not_exists(new_file, 'Cannot preprocess data')
        return VariationFile(self._vcf_file).write(new_file)

    def _do_preprocess_and_persist(self, output_dir: Path) -> SampleData:
        processed_vcf, processed_vcf_index = self._preprocess_vcf(output_dir)
        processed_mask = self._preprocess_mask(output_dir)

        return NucleotideSampleData(sample_name=self.sample_name,
                                    vcf_file=processed_vcf,
                                    vcf_file_index=processed_vcf_index,
                                    mask_bed_file=processed_mask,
                                    preprocessed=True)

    def _do_persist(self, output_dir: Path) -> SampleData:
        if self._mask_bed_file is None:
            raise Exception('mask_bed_file is None')

        new_vcf_file = output_dir / self._vcf_file.name
        new_vcf_index = output_dir / self._vcf_file_index.name
        new_mask_bed_file = output_dir / self._mask_bed_file.name

        self._assert_file_not_exists(new_vcf_file, 'Cannot persist data')
        self._assert_file_not_exists(new_vcf_index, 'Cannot persist data')
        self._assert_file_not_exists(new_mask_bed_file, 'Cannot persist data')

        logger.log(TRACE_LEVEL, f'Copying VCF and BED files to [{output_dir}] for sample [{self.sample_name}]')

        shutil.copy(self._vcf_file, new_vcf_file)
        shutil.copy(self._vcf_file_index, new_vcf_index)
        shutil.copy(self._mask_bed_file, new_mask_bed_file)

        return NucleotideSampleData(sample_name=self.sample_name,
                                    vcf_file=new_vcf_file,
                                    vcf_file_index=new_vcf_index,
                                    mask_bed_file=new_mask_bed_file,
                                    preprocessed=True)

    def is_preprocessed(self) -> bool:
        return self._preprocessed

    def _assert_file_not_exists(self, file: Path, error_prefix: str):
        if file.exists():
            raise Exception(f'{error_prefix} for sample [{self.sample_name}]: file [{file}] exists')

    def read_vcf_features(self) -> pd.DataFrame:
        vcf_file, vcf_file_index = self.get_vcf_file()
        logger.log(TRACE_LEVEL, f'Reading data for sample [{self.sample_name}] from VCF file [{vcf_file}]')
        return VariationFile(vcf_file).read_features(self.sample_name, snpeff_parser=self._snpeff_parser)

    def read_sample_data_features(self, include_masked_regions: bool = True) -> pd.DataFrame:
        vcf_features = self.read_vcf_features()
        vcf_file, vcf_file_index = self.get_vcf_file()

        if include_masked_regions:
            logger.log(TRACE_LEVEL, f'Creating unknown/missing features for sample=[{self.sample_name}]')
            frame_mask = self.get_mask().mask_to_features()
            frame_mask['SAMPLE'] = self.sample_name
            frame_mask['FILE'] = vcf_file.name
            logger.log(TRACE_LEVEL,
                       f'Combining VCF and unknown/missing (mask) dataframes for sample=[{self.sample_name}]')
            frame_vcf_mask = self.combine_vcf_mask(vcf_features, frame_mask)
        else:
            frame_vcf_mask = vcf_features

        return frame_vcf_mask

    def combine_vcf_mask(self, vcf_frame: pd.DataFrame, mask_frame: pd.DataFrame) -> pd.DataFrame:
        """
        Combine features together for VCF variants dataframe with mask dataframe, checking for any overlaps.
        If there is an overlap (e.g., a variant call also is in a masked out region) the masked position (unknown/missing)
        will be preferred as the true feature.
        :param vcf_frame: The dataframe containing only mutations/variant calls.
        :param mask_frame: The dataframe containing features from the genome mask (missing/unknown positions.
        :return: The combined data frame of both types of features.
        """
        combined_df = pd.concat([vcf_frame, mask_frame])

        # Define an order column for TYPE so I can select NUCLEOTIDE_UNKNOWN_TYPE ahead of any other type
        combined_df['TYPE_ORDER'] = 1
        combined_df.loc[combined_df['TYPE'] == NUCLEOTIDE_UNKNOWN_TYPE, 'TYPE_ORDER'] = 0

        # For any overlapping positions, prefer the NUCLEOTIDE_UNKNOWN_TYPE type
        # This may not handle every potential case where a variant overlaps with a masked region
        # (e.g., indel veriants which impact more than one nucleotide) but those should not show up
        # in a VCF file AND also in the mask file if everything was called properly.
        combined_df = combined_df.sort_values(
            ['CHROM', 'POS', 'TYPE_ORDER']).groupby(['CHROM', 'POS'], sort=False).nth(0).reset_index()

        return combined_df.loc[:, self.VCF_FRAME_COLUMNS + self._snpeff_parser.ANNOTATION_COLUMNS]

    def get_vcf_file(self, ignore_preprocessed=False) -> Tuple[Path, Path]:
        if ignore_preprocessed or self._preprocessed:
            return self._vcf_file, self._vcf_file_index
        else:
            raise Exception(f'VCF file for sample [{self.sample_name}] is not preprocessed: {self._vcf_file}')

    def get_mask(self) -> MaskedGenomicRegions:
        return MaskedGenomicRegions.from_file(self.get_mask_file())

    def get_mask_file(self) -> Path:
        if self._preprocessed and self._mask_bed_file is not None:
            return self._mask_bed_file
        else:
            raise Exception(f'Sample mask file is not preprocessed for sample [{self.sample_name}]')



from os import mkdir
from pathlib import Path
from tempfile import TemporaryDirectory

import pytest

from genomics_data_index.storage.model.db.DatabasePathTranslator import DatabasePathTranslator


def test_translate_from_database():
    with TemporaryDirectory() as tmp_dir_str:
        tmp_dir = Path(tmp_dir_str)
        expected1 = tmp_dir / 'file1'
        expected2 = tmp_dir / 'dir' / 'file2'
        expected3 = tmp_dir / 'dir' / 'subdir' / 'file3.txt'

        # Create files
        mkdir(tmp_dir / 'dir')
        mkdir(tmp_dir / 'dir' / 'subdir')
        open(expected1, 'w').close()
        open(expected2, 'w').close()
        open(expected3, 'w').close()

        dpt = DatabasePathTranslator(tmp_dir)

        actual_file = dpt.from_database('file1')
        assert actual_file == expected1

        actual_file = dpt.from_database('dir/file2')
        assert actual_file == expected2

        actual_file = dpt.from_database('dir/subdir/file3.txt')
        assert actual_file == expected3


def test_translate_from_database_fail():
    with TemporaryDirectory() as tmp_dir_str:
        tmp_dir = Path(tmp_dir_str)

        dpt = DatabasePathTranslator(tmp_dir)

        with pytest.raises(Exception) as execinfo:
            dpt.from_database('file1')
        assert 'does not exist for relative path' in str(execinfo.value)


def test_translate_to_database():
    with TemporaryDirectory() as tmp_dir_str:
        tmp_dir = Path(tmp_dir_str)
        file1 = tmp_dir / 'file1'
        file2 = tmp_dir / 'dir' / 'file2'
        file3 = tmp_dir / 'dir' / 'subdir' / 'file3.txt'

        dpt = DatabasePathTranslator(tmp_dir)

        assert 'file1' == dpt.to_database(file1)
        assert 'dir/file2' == dpt.to_database(file2)
        assert 'dir/subdir/file3.txt' == dpt.to_database(file3)



import tempfile
import warnings
from pathlib import Path

import pytest

warnings.filterwarnings("ignore", category=DeprecationWarning)

import genomics_data_index.api as gdi

from genomics_data_index.test.integration import reference_file, sample_dirs, basic_mlst_file
from genomics_data_index.configuration.connector.DataIndexConnection import DataIndexConnection
from genomics_data_index.storage.model.QueryFeatureMutationSPDI import QueryFeatureMutationSPDI
from genomics_data_index.storage.model.QueryFeatureMLST import QueryFeatureMLST
from genomics_data_index.storage.io.mutation.NucleotideSampleDataPackage import NucleotideSampleDataPackage
from genomics_data_index.storage.io.mlst.MLSTSampleDataPackage import MLSTSampleDataPackage
from genomics_data_index.storage.io.mlst.MLSTTSeemannFeaturesReader import MLSTTSeemannFeaturesReader
from genomics_data_index.storage.io.processor.SerialSampleFilesProcessor import SerialSampleFilesProcessor
from genomics_data_index.storage.model.db import Sample


@pytest.fixture
def loaded_genomics_index() -> gdi.GenomicsDataIndex:
    tmp_dir = Path(tempfile.mkdtemp())
    database_connection = DataIndexConnection.connect(database_connection='sqlite:///:memory:',
                                                      database_dir=tmp_dir)

    # Load Nucleotide variation
    database_connection.reference_service.add_reference_genome(reference_file)
    snippy_tmp_dir = Path(tempfile.mkdtemp())
    data_package = NucleotideSampleDataPackage.create_from_snippy(sample_dirs,
                                                                  SerialSampleFilesProcessor(snippy_tmp_dir))
    database_connection.variation_service.insert(data_package, feature_scope_name='genome')

    # Load MLST
    mlst_package = MLSTSampleDataPackage(MLSTTSeemannFeaturesReader(mlst_file=basic_mlst_file))
    database_connection.mlst_service.insert(mlst_package)

    return gdi.GenomicsDataIndex(connection=database_connection)


def test_query_single_mutation(loaded_genomics_index: gdi.GenomicsDataIndex):
    gi = loaded_genomics_index
    db = gi.connection.database
    sampleB = db.get_session().query(Sample).filter(Sample.name == 'SampleB').one()
    sample1 = db.get_session().query(Sample).filter(Sample.name == 'CFSAN002349').one()
    sample2 = db.get_session().query(Sample).filter(Sample.name == 'CFSAN023463').one()

    query_result = gi.samples_query().hasa(QueryFeatureMutationSPDI('reference:5061:G:A'))
    assert 1 == len(query_result)
    assert {sampleB.id} == set(query_result.sample_set)
    assert 9 == len(query_result.universe_set)

    query_result = gi.samples_query().hasa(QueryFeatureMutationSPDI('reference:5061:G:A'))
    assert 1 == len(query_result)
    assert {sampleB.id} == set(query_result.sample_set)
    assert 9 == len(query_result.universe_set)

    query_result = gi.samples_query().hasa(QueryFeatureMLST('mlst:lmonocytogenes:abcZ:1'))
    assert 2 == len(query_result)
    assert {sample1.id, sample2.id} == set(query_result.sample_set)
    assert 9 == len(query_result.universe_set)

    assert {'CFSAN002349', 'CFSAN023463'} == set(query_result.tolist())
    assert {sample1.id, sample2.id} == set(query_result.tolist(names=False))



import json
import logging
import re
from collections import Counter
from pathlib import Path
from typing import List, Dict, Set, Any, Union

from Bio.SeqRecord import SeqRecord

from genomics_data_index.api.query.SamplesQuery import SamplesQuery
from genomics_data_index.api.query.kind.isa.typing.SamplesTypingIsaKind import SamplesTypingIsaKind
from genomics_data_index.storage.SampleSet import SampleSet

logger = logging.getLogger(__name__)


class MutationParsingError(Exception):

    def __init__(self, msg):
        super().__init__(msg)


class ExperimentalSARSCov2ConstellationsTyper(SamplesTypingIsaKind):
    """
    This is an experimental/test class to try out typing SARS-CoV-2 based on constellations of variants/mutations.
    This parses constellation definitions from https://github.com/cov-lineages/constellations/blob/main/constellations/definitions
    TODO: I need to review code in here https://github.com/cov-lineages/scorpio/blob/main/scorpio/scripts/type_constellations.py
          to make sure I match what's going on there.
    TODO: There are still some bugs with this. In particular, I've noticed typing B.1.1.7 doesn't work due to different standards in naming
          mutations between HGVS notation and names given in the constellation files (S:p.Y145del vs. s:Y144-).
    """

    def __init__(self, constellation_files: Union[List[Path], List[str]],
                 sequence: SeqRecord, perfect_match_only: bool = False,
                 valid_gene_identifiers: Set[str] = None):
        super().__init__()

        sequence_name = str(sequence.id)
        if sequence_name not in ['NC_045512.2', 'NC_045512']:
            logger.warning(
                f"sequence_name=[{sequence_name}] is not 'NC_045512.2' or 'NC_045512'. This is what the constellations are with respect"
                "to <https://github.com/cov-lineages/constellations/blob/main/constellations/data/SARS-CoV-2.json>")

        self._sequence_name = sequence_name
        self._sequence = sequence
        if valid_gene_identifiers is None:
            self._valid_gene_identifiers = {'ORF1ab', 'S', 'ORF3a', 'E', 'M', 'ORF6', 'ORF7a', 'ORF7b', 'ORF8', 'N',
                                            'ORF10'}
        self._typing_definitions = self._parse_definitions(constellation_files)
        self._perfect_match_only = perfect_match_only

    def _handle_deletion_identifier(self, mutation_values: List[str], mutation: str) -> str:
        if mutation_values[0] == 'del':
            # Based on looking at some real-world data (for a single deletion identifier), it looks like the
            # position here is the position of the unchanged base, with the deletion occuring right after.
            position = int(mutation_values[1])
            position_0coord = position - 1
            length = int(mutation_values[2])

            if position_0coord < 0:
                raise MutationParsingError(f'Cannot parse mutation [{mutation}], position is negative')
            else:
                # I need to specify an insertion that's not of length 0 (for my use of these identifiers).
                # For example, say the sequence is "ATCG" and position=1 (position right before deletion),
                # and deletion length = 2. Then this means deleting "A[TC]G" and inserting nothing.
                # But to convert this to an identifier where I have at least 1 insertion, I need to delete
                # and re-insert one of the reference characters. I do this with the left-most base, so the actual
                # deletion becomes "[ATC]G" followed by an insertion of "A". That is, the identifier
                # becomes "sequence:ATC:A".
                deletion_start = position_0coord
                deletion_stop = deletion_start + (length + 1)
                position_start = position

            deletion_sequence = self._sequence.seq[deletion_start:deletion_stop]
            insertion_sequence = self._sequence.seq[deletion_start:deletion_start + 1]

            return f'{self._sequence_name}:{position_start}:{deletion_sequence}:{insertion_sequence}'
        else:
            raise MutationParsingError(f'Cannot parse mutation [{mutation}], has three parts but first '
                                       f'is not "del".')

    def _handle_gene_nucleotide_identifier(self, mutation_values: List[str], mutation: str) -> str:
        gene, mutation = mutation_values
        if gene == 'nuc':
            match = re.match(r'([ATCG]+)(\d+)([ATCG]+)', mutation)
            if not match:
                raise MutationParsingError(f'Could not parse mutation={mutation}, does not match pattern like [A150T].')
            mutation_id = f'{self._sequence_name}:{match.group(2)}:{match.group(1)}:{match.group(3)}'
        else:
            # Curation of different possible gene identifiers
            gene_curation_map = {
                's': 'S',
                '1ab': 'ORF1ab',
                'orf1ab': 'ORF1ab',
                'ORF1a': 'ORF1ab',
                'ORF1b': 'ORF1ab',
                'NSP2': 'ORF1ab',
                'nsp3': 'ORF1ab',
                'nsp4': 'ORF1ab',
                'nsp5': 'ORF1ab',
                'nsp6': 'ORF1ab',
                'nsp7': 'ORF1ab',
                'nsp12': 'ORF1ab',
                'nsp13': 'ORF1ab',
                'nsp15': 'ORF1ab',
                '8': 'ORF8'
            }
            if gene in gene_curation_map:
                gene = gene_curation_map[gene]
            if gene not in self._valid_gene_identifiers:
                raise MutationParsingError(
                    f'gene=[{gene}] for mutation=[{mutation}] is not one of the valid '
                    f"gene identifiers={self._valid_gene_identifiers}.")
            if mutation.endswith('-'):
                mutation = self._convert_aa_deletion(mutation)
            mutation_id = f'hgvs_gn:{self._sequence_name}:{gene}:p.{mutation}'
        return mutation_id

    def mutation_to_identifier(self, mutation: str) -> str:
        values = mutation.split(':')
        if len(values) == 2:
            return self._handle_gene_nucleotide_identifier(values, mutation=mutation)
        elif len(values) == 3:
            return self._handle_deletion_identifier(values, mutation=mutation)
        else:
            raise MutationParsingError(f'Cannot parse mutation [{mutation}], cannot be split by ":".')

    def _convert_aa_deletion(self, mutation: str) -> str:
        match = re.match(r'(\D+)(\d+)-$', mutation)
        if not match:
            raise MutationParsingError(f'Could not parse mutation={mutation}, does not match pattern like [HV69-].')
        aa_values = match.group(1)
        position = match.group(2)
        if len(aa_values) == 1:
            return f'{aa_values}{position}del'
        else:
            first_aa = aa_values[0]
            last_aa = aa_values[-1]
            last_position = int(position) + len(aa_values) - 1
            return f'{first_aa}{position}_{last_aa}{last_position}del'

    def _parse_definitions(self, constellation_files: Union[List[Path], List[str]]) -> Dict[str, Any]:
        typing_definitions = dict()
        for file in constellation_files:
            if isinstance(file, str):
                file = Path(file)

            try:
                constellation_info = self._load_definitions(file)
                signature_mutations_raw = constellation_info['sites']
                signature_mutation_ids = {self.mutation_to_identifier(x) for x in signature_mutations_raw}
                logger.debug(f'constellation_file=[{file}], mutation_ids={signature_mutation_ids}')

                rules = constellation_info['rules']
                min_alt = rules.get('min_alt', len(signature_mutation_ids))
                max_ref = rules.get('max_ref', 0)
                other_rules = set(rules.keys()) - {'min_alt', 'max_ref'}

                # Handle any other special-rules
                must_have_mutations = set()
                for rule_key in other_rules:
                    if rules[rule_key] == 'alt':
                        mutation_id = self.mutation_to_identifier(rule_key)
                        must_have_mutations.add(mutation_id)
                    else:
                        logger.warning(f'Skipping unknown rule {rule_key}={rules[rule_key]}, file=[{file.name}]')

                labels = constellation_info['tags']
                labels = labels + [constellation_info['label']]

                for label in labels:
                    if label in typing_definitions:
                        previous_definition = typing_definitions[label]
                        previous_file = previous_definition['file']
                        logger.warning(f'Attempting to set typing definition for tag/label=[{label}], '
                                       f'file=[{file.name}], '
                                       f'but it has already been set in file [{previous_file.name}]. '
                                       f'Will ignore definition for tag/label=[{label}], file=[{file.name}].')
                    else:
                        typing_definitions[label] = {
                            'signature_mutations': signature_mutation_ids,
                            'signature_mutations_original': signature_mutations_raw,
                            'must_have_mutations': must_have_mutations,
                            'min_alt': min_alt,
                            'max_ref': max_ref,
                            'file': file,
                            'tags': labels,
                        }
            except MutationParsingError as e:
                logger.error(f'Skipping file [{file}]: {e}')
        return typing_definitions

    def _load_definitions(self, constellation_file: Path) -> Dict[str, Any]:
        with open(constellation_file, 'r') as f:
            data = json.load(f)
        return data

    @property
    def name(self) -> str:
        return 'constellations_typer'

    @property
    def version(self) -> str:
        return 'experimental'

    def _validate_type_name(self, type_name: str) -> None:
        if type_name not in self._typing_definitions:
            raise Exception(f'type_name=[{type_name}] is not a valid type.')

    @property
    def type_names(self) -> List[str]:
        return list(self._typing_definitions.keys())

    def perfect_matches(self, type_name: str, query: SamplesQuery,
                        signature_mutations: Set[str] = None) -> SamplesQuery:
        self._validate_type_name(type_name)
        if signature_mutations is None:
            signature_mutations = self._typing_definitions[type_name]['signature_mutations']

        q = query
        for mutation in signature_mutations:
            q = q.hasa(mutation, kind='mutation')

        return q

    def _imperfect_pass_criteria(self, mutation_present_count: int, mutation_unknown_count: int,
                                 total_mutations: int,
                                 min_alt: int, max_ref: int) -> bool:
        number_ref = total_mutations - mutation_present_count - mutation_unknown_count
        return (mutation_present_count >= min_alt) and (number_ref <= max_ref)

    def imperfect_matches(self, data: str, query: SamplesQuery) -> SamplesQuery:
        self._validate_type_name(data)
        signature_mutations = self._typing_definitions[data]['signature_mutations']
        total_mutations = len(signature_mutations)
        min_alt = self._typing_definitions[data]['min_alt']
        max_ref = self._typing_definitions[data]['max_ref']

        # Use a Counter to keep track of number of mutations associated with each sample
        samples_present_counter = Counter(query.sample_set)
        sample_unknown_counter = Counter(query.unknown_set)

        for mutation in signature_mutations:
            mutation_query = query.hasa(mutation, kind='mutation')
            samples_present_counter.update(mutation_query.sample_set)
            sample_unknown_counter.update(mutation_query.unknown_set)

        sample_ids_pass = {s for s in samples_present_counter if
                           self._imperfect_pass_criteria(samples_present_counter[s],
                                                         sample_unknown_counter[s],
                                                         total_mutations=total_mutations,
                                                         min_alt=min_alt, max_ref=max_ref)}
        sample_set_present = SampleSet(sample_ids_pass)

        return query.intersect(sample_set_present)

    def isa_type(self, data: str, query: SamplesQuery) -> SamplesQuery:
        self._validate_type_name(data)
        signature_mutations = self._typing_definitions[data]['signature_mutations']
        must_have_mutations = self._typing_definitions[data]['must_have_mutations']
        min_alt = self._typing_definitions[data]['min_alt']

        # If any must_have mutations, handle these first to try to eliminate as much as possible ahead of time
        if len(must_have_mutations) > 0:
            query = self.perfect_matches(type_name=data, query=query, signature_mutations=must_have_mutations)

        if self._perfect_match_only or min_alt == len(signature_mutations):
            # Separate case for all perfect matches since I can take advantage of tracking unknowns
            return self.perfect_matches(type_name=data, query=query)
        else:
            # SARS-CoV-2 typing with constellations does not have the concept of an "unknown" sample and
            # instead tries to classify these using e.g., min_alt and max_ref, and so on
            # So, for the purposes of typing, I want to ignore any "unknown" samples for typing and add
            # the unknown set back in at the end. The is because no matter what the result of the sequence-typing
            # code below, the statement 'query.isa("Type")' will still be unknown if the sample was already unknown
            # in the input query (for example, if Sample A is already "Unknown" in the input query and the result
            # below is "True" for Sample A, then the final result for Sample A is still "Unknown" since
            # "Unknown" AND "True" = "Unknown").
            query_no_unknowns = query.select_present().reset_universe()

            # Look for perfect matches
            query_perfect_match = self.perfect_matches(type_name=data,
                                                       query=query_no_unknowns,
                                                       signature_mutations=signature_mutations)

            # Look for imperfect matches in what's left over
            query_imperfect_matches = query_perfect_match.select_unknown() | query_perfect_match.select_absent()
            query_imperfect_matches = self.imperfect_matches(data, query_imperfect_matches)

            # and query to add back any existing unknown samples
            return query & (query_perfect_match.select_present() | query_imperfect_matches.select_present())



import abc
from typing import Tuple

from ete3 import Tree

from genomics_data_index.storage.SampleSet import SampleSet


class TreeBuilder(abc.ABC):

    def __init__(self):
        pass

    @abc.abstractmethod
    def build(self, samples_set: SampleSet, method: str, **kwargs) -> Tuple[Tree, int, SampleSet]:
        pass
